# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.
"""BandSet manager.

Core class that manages BandSets.
A BandSet is an object that includes information about single bands
(from the file path to the spatial and spectral characteristics).
Bands in a BandSet can be referenced by the properties thereof,
such as order number or center wavelength.

BandSets can be used as input for operations on multiple bands
such as Principal Components Analysis, classification, mosaic,
or band calculation.
Multimple BandSets can be defined and identified by their reference number
in the BandSet Catalog.

Most BandSet functions can be accessed through the BandSet Catalog.

Typical usage example:

    >>> # import Remotior Sensus and start the session
    >>> import remotior_sensus
    >>> rs = remotior_sensus.Session()
    >>> # set lists of files, names and center wavelength
    >>> file_list = ['file_1.tif', 'file_2.tif', 'file_3.tif']
    >>> band_names = ['name_1', 'name_2', 'name_3']
    >>> wavelengths = [0.6, 0.7, 0.8]
    >>> # set optional date
    >>> date = '2020-01-01'
    >>> # create a BandSet
    >>> bandset = rs.bandset.create(
    ... file_list, band_names=band_names, wavelengths=wavelengths, dates=date
    ... )
    >>> # get list of absolute paths of bands
    >>> absolute_paths = bandset.get_absolute_paths()
    >>> # bandset wavelengths
    >>> wavelengths = bandset.get_wavelengths()
    >>> # create a BandSet with a root directory for files in relative paths
    >>> root_directory = 'root_directory'
    >>> relative_file_list = ['file_1.tif', 'file_2.tif', 'file_3.tif']
    >>> bandset = rs.bandset.create(file_list,root_directory=root_directory)
    >>> # get list of relative paths of bands
    >>> relative_paths = bandset.get_paths()
    >>> # get band by nearest wavelength
    >>> band_x = bandset.get_band_by_wavelength(0.8,threshold=0.1)
"""

import random
from typing import Union, Optional

import numpy as np

from remotior_sensus.core import (
    configurations as cfg, messages, table_manager as tm
)
from remotior_sensus.util import dates_times, files_directories, raster_vector


class BandSet(object):
    """Manages band sets.

    This module allows for managing bands in a BandSet.

    Attributes:
        bands: BandSet of band tables.
        crs: BandSet coordinate reference system.
        get: alias for get_band function.
        
    Examples:
        Create a BandSet
            >>> import remotior_sensus
            >>> rs = remotior_sensus.Session()
            >>> file_list = ['file_1.tif', 'file_2.tif', 'file_3.tif']
            >>> bandset = rs.bandset.create(file_list)
            
        Get the first band
            >>> band_1 = bandset.get_band(1)
    """  # noqa: E501

    def __init__(
            self, bandset_uid, bands_list=None, name=None, date=None,
            root_directory=None, crs=None, box_coordinate_list=None
    ):
        """Initializes a BandSet.

        Initializes a BandSet defining a unique ID.
        One should use the create method for creating new BandSets.       

        Args:
            bandset_uid: unique BandSet ID.
            bands_list: list of band tables.
            name: optional BandSet name.
            date: optional BandSet date.
            root_directory: optional BandSet root directory for relative path.
            crs: BandSet coordinate reference system.
            box_coordinate_list: list of coordinates [left, top, right, bottom] 
                to create a virtual subset.

        Examples:
            Initialize a BandSet
                >>> BandSet.create(name=name,box_coordinate_list=box_coordinate_list)
        """  # noqa: E501
        self.bands = tm.create_bandset_table(bands_list)
        # unique ID
        self.uid = bandset_uid
        self.date = date
        self.root_directory = root_directory
        self.crs = crs
        # alias for get_band
        self.get = self.get_band
        if name is None:
            names = self.get_band_attributes('name')
            try:
                self.name = names[0][:-2]
            except Exception as err:
                str(err)
                if names is not None:
                    if len(names) > 0:
                        self.name = names[0]
                    else:
                        self.name = 'BandSet'
                else:
                    self.name = 'BandSet'
        else:
            self.name = name
        self.box_coordinate_list = box_coordinate_list

    @property
    def date(self):
        """Optional date."""
        return self._date

    @date.setter
    def date(self, date):
        self._date = date

    @property
    def uid(self):
        """str: Unique BandSet ID."""
        return self._uid

    @uid.setter
    def uid(self, uid):
        self._uid = uid

    @property
    def root_directory(self):
        """str: Optional BandSet root directory for relative path."""
        return self._root_directory

    @root_directory.setter
    def root_directory(self, root_directory):
        self._root_directory = root_directory

    @property
    def name(self):
        """str: Optional BandSet name."""
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def box_coordinate_list(self):
        """
        list: Optional BandSet box coordinate list for virtual subset.
        """  # noqa: E501
        return self._box_coordinate_list

    @box_coordinate_list.setter
    def box_coordinate_list(self, box_coordinate_list):
        self._box_coordinate_list = box_coordinate_list

    def execute(self, function, *args, **kwargs):
        """Executes a function.

        Executes a functions directly from the BandSet, passing the argument input_bands.
        The arguments are related to the function.

        Args:
            function: the function to be executed

        Returns:
            the output of the function

        Examples:
            Given that a session was previously started
                >>> import remotior_sensus
                >>> rs = remotior_sensus.Session()
                >>> bandset = rs.bandset.create(
                ...     ['file1.tif', 'file2.tif'], wavelengths=['Sentinel-2'],
                ... )
                
            Calculation of sum of the first two bands 
                >>> output_object = bandset.execute(
                ...     rs.band_calc, output_path='output.tif', expression_string='"b1" + "b2"'
                ... )
                
            Calculation of NDVI
                >>> output_object = bandset.execute(
                ...     rs.band_calc, output_path='output.tif', 
                ...     expression_string='("#NIR#" - "#RED#") / ("#NIR#" + "#RED#")'
                ... )
                
            Calculation of band combination
                >>> output_object = bandset.execute(rs.band_combination, output_path='output.tif')

        """  # noqa: E501

        kwargs['input_bands'] = self
        return function(*args, **kwargs)

    def get_band_count(self) -> int:
        """Gets the count of bands."""
        if self.bands is None:
            return 0
        else:
            return self.bands.shape[0]

    def get_paths(self) -> list:
        """Gets the list of bands."""
        return self.bands['path'].tolist()

    def get_absolute_paths(self) -> list:
        """Gets the list of absolute paths."""
        paths = []
        if self.bands is not None:
            paths = self.bands['absolute_path'].tolist()
        cfg.logger.log.debug('absolute_paths: %s' % str(paths))
        return paths

    def get_wavelengths(self) -> list:
        """Gets the list of center wavelength."""
        return self.bands['wavelength'].tolist()

    def get_wavelength_units(self) -> list:
        """Gets the list of wavelength units."""
        return self.bands['wavelength_unit'].tolist()

    def get_raster_band_list(self) -> list:
        """Gets the list of raster bands."""
        raster_band = self.bands['raster_band'].tolist()
        band_list = []
        for n in raster_band:
            band_list.append([n])
        return band_list

    def get_band_attributes(self, attribute: str = None) -> list:
        """Gets band attributes.

        Gets an attribute of bands.

        Args:
            attribute: attribute name.
            
        Returns:
            returns list of attributes; if attribute is None returns the bands;  
            if attribute is not found returns None.

        Examples:
            Get band names
            >>> bandset = BandSet()
            >>> names = bandset.get_band_attributes('name')
        """  # noqa: E501
        cfg.logger.log.debug('attribute: %s' % str(attribute))
        if attribute is None:
            result = self.bands
            cfg.logger.log.debug('get bands')
        else:
            if self.bands is not None:
                record = self.bands[attribute]
                if len(record) == 0:
                    result = None
                else:
                    result = record.tolist()
            else:
                result = None
        return result

    def get_bands_by_attributes(
            self, attribute: str, attribute_value, output_as_number=False
    ) -> np.array:
        """Gets bands by attribute.

        Gets bands identified by an attribute value (identified bands have 
        the attribute value equal to the attribute_value).

        Args:
            attribute: attribute name.
            attribute_value: attribute value.
            output_as_number: if True returns the number of the band, if False returns the band table.
            
        Returns:
            if output_number is True, returns list of band number;
            if output_number is False, returns bands identified by attribute;
            returns None if no band has the attribute value.

        Examples:
            Get bands by attributes
            >>> bandset = BandSet()
            >>> band_x = bandset.get_bands_by_attributes('wavelength',0.8)
        """  # noqa: E501
        cfg.logger.log.debug(
            'attribute: %s; attribute_value: %s; output_number: %s' % (
                attribute, attribute_value, output_as_number)
        )
        result = None
        if self.bands is not None:
            bands = self.bands[self.bands[attribute] == attribute_value]
            if not output_as_number:
                result = bands
            else:
                result = bands['band_number'].tolist()
        return result

    def get_band_alias(self) -> list:
        """Gets bands alias (with band number) used in band_calc

        Returns:
            the list of band aliases.
        """
        result = []
        if self.bands is not None:
            for i in range(1, len(self.bands) + 1):
                result.append('%s%s' % (cfg.variable_band_name, i))
        cfg.logger.log.debug('result: %s' % str(result))
        return result

    def get_band(self, number=None) -> np.array:
        """Gets a band.

        Gets a band by number.

        Args:
            number: band number.

        Returns:
            the band table; if number is None returns all the bands;
            returns None if band number is not found.

        Examples:
            Get bands by number
            >>> bandset = BandSet()
            >>> # get first band
            >>> band = bandset.get_band(1)
            >>> # band x size
            >>> band_x_size = band.x_size
            >>> # get band nodata directly
            >>> band_nodata = bandset.get_band(1).nodata
        """  # noqa: E501
        cfg.logger.log.debug('number: %s' % number)
        if number is None:
            result = self.bands
        elif not number:
            result = None
        else:
            try:
                result = self.bands[self.bands['band_number'] == number]
            except Exception as err:
                str(err)
                result = None
        return result

    def reset(self):
        """Resets a BandSet.

        Resets a BandSet destroyng bands and attributes.

        Examples:
            Reset a BandSet
            >>> bandset = BandSet()
            >>> # reset
            >>> bandset.reset()
        """  # noqa: E501
        self.bands = self.name = self.date = self.root_directory = None
        self.box_coordinate_list = self.crs = None
        cfg.logger.log.debug('reset')

    @classmethod
    def create(
            cls, paths: Union[list, str] = None,
            band_names: Optional[list] = None,
            wavelengths: Optional[list or str] = None,
            unit: Optional[str] = None,
            multiplicative_factors: Optional[list] = None,
            additive_factors: Optional[list] = None,
            dates: Optional[list or str] = None,
            root_directory: Optional[str] = None,
            box_coordinate_list: Optional[list] = None,
            name: Optional[str] = None
    ):
        """Creates a BandSet.

        This method creates a BandSet defined by input files.
        Raster properties are derived from files to populate 
        the table of bands.

        Args:
            name: BandSet name.
            paths: list of file paths or a string of directory path; also, a list of directory path 
                and name filter is accepted.
            band_names: list of raster names used for identifying the bands, 
                if None then the names are automatically extracted from file names.
            wavelengths: list of center wavelengths of bands or string of sensor names (also partial).
            unit: wavelength unit.
            multiplicative_factors: multiplicative factors for bands during calculations.
            additive_factors: additive factors for bands during calculations.
            dates: list of date strings, or single date string (format YYYY-MM-DD) 
                or string defined in configurations date_auto to detect date from directory name.
            root_directory: root directory for relative path.
            box_coordinate_list: list of coordinates [left, top, right, bottom] 
                to create a virtual subset.

        Returns:
            returns BandSet.
            
        Examples:
            Given that a session was previously started
                >>> import remotior_sensus
                >>> rs = remotior_sensus.Session()
                >>> # set lists of files, names and center wavelength
                >>> file = ['file_1.tif', 'file_2.tif', 'file_3.tif']
                >>> names = ['name_1', 'name_2', 'name_3']
                >>> wavelength = [0.6, 0.7, 0.8]
                >>> # create a BandSet
                >>> bandset = rs.bandset.create(file,band_names=names,wavelengths=wavelength)
                
           Passing a directory with file name filter and the wavelenght from satellite name.
                >>> bandset = rs.bandset.create(['directory_path', 'tif'], wavelengths='Sentinel-2')
        """  # noqa: E501
        cfg.logger.log.info('start')
        bandset_uid = cls.generate_uid()
        satellite = None
        multiband = None
        # directory
        if paths is None:
            cfg.logger.log.info('end')
            return cls(bandset_uid=bandset_uid, name=name)
        elif len(paths) <= 2 and files_directories.is_directory(
                files_directories.relative_to_absolute_path(
                    paths[0], root_directory
                )
        ):
            filters = None
            try:
                if len(paths) == 2:
                    filters = paths[1].strip("'").strip()
            except Exception as err:
                cfg.logger.log.error(str(err))
            file_list = files_directories.files_in_directory(
                paths[0], sort_files=True, path_filter=filters,
                root_directory=root_directory
            )
            # date from directory name
            if dates is not None and type(dates) is not list:
                if dates.lower() == cfg.date_auto:
                    dates = dates_times.date_string_from_directory_name(
                        paths[0]
                    )
        elif type(paths) is str and files_directories.is_directory(
                files_directories.relative_to_absolute_path(
                    paths, root_directory
                )
        ):
            file_list = files_directories.files_in_directory(
                paths, sort_files=True, root_directory=root_directory
            )
            # date from directory name
            if dates is not None and type(dates) is not list:
                if dates.lower() == cfg.date_auto:
                    dates = dates_times.date_string_from_directory_name(
                        paths[0]
                    )
        # one file
        elif len(paths) == 1:
            b_count = raster_vector.get_number_bands(paths[0])
            if b_count > 1:
                multiband = b_count
            file_list = paths.copy()
        # single band rasters
        else:
            file_list = paths.copy()
        # wavelength
        try:
            if wavelengths is not None and len(wavelengths) == 1:
                for sat in cfg.satWlList:
                    if (wavelengths[0].lower() in sat.lower()
                            or wavelengths[0].lower().replace(' ', '')
                            in sat.lower().replace(' ', '')):
                        satellite = sat
                        wavelengths = None
                        break
        except Exception as err:
            cfg.logger.log.error(str(err))
            wavelengths = None
        if unit is None:
            unit = cfg.no_unit
        if satellite is not None and satellite != cfg.no_satellite:
            sat_wl, sat_unit, sat_bands = cfg.satellites[satellite]
        else:
            sat_wl = sat_unit = sat_bands = None
        # create band list
        bands_list = []
        date = None
        counter = 1
        for f in range(len(file_list)):
            # band names from raster
            if band_names is None:
                band_name = _raster_to_band_names(file_list[f])[0]
            else:
                band_name = band_names[f]
            # path
            path = file_list[f]
            cfg.logger.log.debug('band %s path: %s' % (str(counter), path))
            # date
            if dates is not None:
                if type(dates) is not list:
                    if dates.lower() == cfg.date_auto:
                        date = None
                    else:
                        date = dates
                else:
                    date = dates[f]
            else:
                date = None
            # get wavelength
            if wavelengths is None:
                if satellite is not None and satellite != cfg.no_satellite:
                    wl = sat_wl[f]
                    unit = sat_unit
                    # get band number from names
                    try:
                        # numbers in format 01, 02, ...
                        b = band_name.lower()[-2:]
                        wl = float(sat_wl[sat_bands.index(b)])
                    except Exception as err:
                        e = str(err)
                        try:
                            # numbers in format 1, 2, ...
                            b = band_name.lower()[-2:].lstrip('0')
                            wl = float(sat_wl[sat_bands.index(b)])
                        except Exception as err:
                            e = '{}; {}'.format(err, e)
                            # get values from list
                            try:
                                wl = float(sat_wl[f])
                            except Exception as err:
                                cfg.logger.log.error('%s: %s', (err, e))
                else:
                    wl = f + 1
                    unit = cfg.no_unit
            else:
                wl = wavelengths[f]
            # single band
            if multiband is None:
                # multiplicative factors
                if multiplicative_factors is None:
                    multiplicative_factors = [1] * len(file_list)
                # additive factors
                if additive_factors is None:
                    additive_factors = [0] * len(file_list)
                new_band = _create_table_of_bands(
                    path, band_number=counter, raster_band=1, name=band_name,
                    multiplicative_factor=multiplicative_factors[f],
                    additive_factor=additive_factors[f], date=date,
                    root_directory=root_directory, wavelength=wl,
                    wavelength_unit=unit
                )
                bands_list.append(new_band)
                counter += 1
            # multi band raster
            else:
                # multiplicative factors
                if multiplicative_factors is None:
                    multiplicative_factors = [1] * multiband
                # additive factors
                if additive_factors is None:
                    additive_factors = [0] * multiband
                # band names from raster
                band_names = _raster_to_band_names(file_list[0], multiband)
                for b in range(multiband):
                    new_band = _create_table_of_bands(
                        path, band_number=counter, raster_band=b + 1,
                        name=band_names[b],
                        multiplicative_factor=multiplicative_factors[b],
                        additive_factor=additive_factors[b], date=date,
                        root_directory=root_directory, wavelength=wl,
                        wavelength_unit=unit
                    )
                    bands_list.append(new_band)
                    counter += 1
        if len(bands_list) > 0:
            crs = bands_list[0]['crs'][0]
        else:
            crs = None
        cfg.logger.log.info('end; file list: %s' % file_list)
        return cls(
            bands_list=bands_list, name=name, date=date,
            bandset_uid=bandset_uid, crs=crs,
            box_coordinate_list=box_coordinate_list
        )

    def get_band_by_wavelength(
            self, wavelength: float, threshold: float = None,
            output_as_number=False
    ) -> np.array:
        """Gets band by wavelength.

        Gets band by nearest wavelength.

        Args:
            wavelength: value of wavelength.
            threshold: threshold for wavelength identification, identifying bands within wavelength +- threshold; if None, the nearest value is returned.
            output_as_number: if True returns the number of the band, if False returns the band table.

        Returns:
            returns band table.
            
        Examples:
            Get band by wavelength
                >>> import remotior_sensus
                >>> rs = remotior_sensus.Session()
                >>> bandset = BandSet()
                >>> # get band by nearest wavelength
                >>> band_x = bandset.get_band_by_wavelength(
                >>> wavelength=0.7, threshold=0.1
                    )
        """  # noqa: E501
        try:
            wl = float(wavelength)
        except Exception as err:
            cfg.logger.log.error(str(err))
            return None
        band_table = tm.find_nearest_value(
            self.bands, 'wavelength', wl, threshold
        )
        if not output_as_number and output_as_number is not None:
            band = band_table
        elif band_table is not None:
            band = band_table['band_number']
        else:
            band = None
        cfg.logger.log.debug(
            'wavelength %s; band: %s'
            % (str(wavelength), str(band))
        )
        return band

    def spectral_range_bands(self, output_as_number=True) -> list:
        """Gets bands from spectral range.

        Gets bands from spectral range blue, green, red, nir, swir_1, swir_2.

        Args:
            output_as_number: if True returns the number of the band, if False returns the band table.

        Returns:
            returns list of bands.
            
        Examples:
            Gets bands from spectral range
                >>> import remotior_sensus
                >>> rs = remotior_sensus.Session()
                >>> bandset = BandSet()
                >>> # spectral range bands
                >>> (blue_band, green_band, red_band, nir_band, swir_1_band,
                ... swir_2_band) = bandset.spectral_range_bands(
                ... output_as_number=False)
        """  # noqa: E501
        cfg.logger.log.debug('start')
        try:
            unit = self.get_wavelength_units()[0]
        except Exception as err:
            str(err)
            return [None, None, None, None, None, None]
        # scale unit
        if unit == cfg.wl_micro:
            m = 1
        elif unit == cfg.wl_nano:
            m = 1000
        else:
            m = 0
        # centers and threshold list blue, green, red, nir, swir_1, swir_2
        c_list = [[cfg.blue_center * m, cfg.blue_threshold * m],
                  [cfg.green_center * m, cfg.green_threshold * m],
                  [cfg.red_center * m, cfg.red_threshold * m],
                  [cfg.nir_center * m, cfg.nir_threshold * m],
                  [cfg.swir_1_center * m, cfg.swir_1_threshold * m],
                  [cfg.swir_2_center * m, cfg.swir_2_threshold * m]]
        # get bands
        bands = []
        for c in c_list:
            band = self.get_band_by_wavelength(
                wavelength=c[0], threshold=c[1],
                output_as_number=output_as_number
            )
            bands.append(band)
        cfg.logger.log.debug('end; bands: %s' % str(bands))
        return bands

    def add_new_band(
            self, path: str, band_number: int = None, raster_band=None,
            band_name=None, date=None, root_directory=None,
            multiplicative_factor=None, additive_factor=None, wavelength=None,
            unit=None
    ):
        """Adds new band to a BandSet.

        Adds a new band to a BandSet with the option to set the band number
        for the position in the BandSet order.

        Args:
            path: file path.
            band_number: sets the position of the band in BandSet order.
            raster_band: number of band for multiband rasters.
            band_name: name of raster used for band.
            date: single date (as YYYY-MM-DD).
            multiplicative_factor: multiplicative factor.
            additive_factor: additive factor.
            wavelength: center wavelength.
            unit: wavelength unit.
            root_directory: root directory for relative path.

        Examples:
            Add a new band
                >>> import remotior_sensus
                >>> rs = remotior_sensus.Session()
                >>> bandset = BandSet()
                >>> bandset.add_new_band(
                ... path=path, band_number=band_number, raster_band=raster_band,
                ... band_name=band_name, date=date, root_directory=root_directory,
                ... multiplicative_factor=multiplicative_factor,
                ... additive_factor=additive_factor, wavelength=wavelength, unit=unit
                ... )
        """  # noqa: E501
        if band_number is None:
            band_number = self.get_band_count() + 1
        if band_name is None:
            band_name = _raster_to_band_names(path)[0]
        if self.root_directory != root_directory:
            cfg.logger.log.warning('root_directory: %s' % root_directory)
            messages.warning('root_directory: %s' % root_directory)
        if multiplicative_factor is None:
            multiplicative_factor = 1
        if additive_factor is None:
            additive_factor = 0
        if wavelength is None:
            wavelength = band_number
            unit = cfg.no_unit
        if unit is None:
            unit = cfg.no_unit
        # move up bands above added band
        if band_number <= self.get_band_count():
            self.bands['band_number'][
                self.bands['band_number'] >= band_number] = \
                self.bands['band_number'][
                    self.bands['band_number'] >= band_number] + 1
        band = _create_table_of_bands(
            path, band_number=band_number, raster_band=raster_band,
            name=band_name, multiplicative_factor=multiplicative_factor,
            additive_factor=additive_factor, date=date,
            root_directory=root_directory, wavelength=wavelength,
            wavelength_unit=unit
        )
        if band is not None:
            if self.bands is None:
                self.bands = tm.create_bandset_table([band])
            else:
                self.bands = tm.create_bandset_table([self.bands, band])

    def sort_bands_by_wavelength(self):
        """Sorts band order by wavelength"""
        self.bands.sort(order='wavelength')
        order_list = list(range(1, self.bands.shape[0] + 1))
        self.bands['band_number'] = np.array(order_list)

    def find_values_in_list(
            self, attribute, value_list, output_attribute=None
    ) -> list:
        """Adds new band to a BandSet.

        Finds BandSet values in a list and return a band attribute or band numbers.

        Args:
            attribute: attribute name for identification.
            value_list: attribute value list for indentification.
            output_attribute: attribute name of desired output; if None returns the band number.

        Returns:
            returns list of band attributes.
            
        Examples:
            Find values from wavelength attribute
                >>> import remotior_sensus
                >>> rs = remotior_sensus.Session()
                >>> bandset = BandSet()
                >>> bandset_values = bandset.find_values_in_list(
                ... attribute='wavelength', value_list=[0.6, 0.7],
                ... output_attribute='path'
                ... )
        """  # noqa: E501
        band_list = []
        for value in value_list:
            if output_attribute is None:
                bandset_attributes = self.bands['band_number'][
                    self.bands[attribute] == value]
            else:
                bandset_attributes = self.bands[output_attribute][
                    self.bands[attribute] == value]
            if len(bandset_attributes) > 0:
                for attr in bandset_attributes:
                    band_list.append(attr)
        return band_list

    @staticmethod
    def generate_uid() -> str:
        """Generates unique ID for BandSet

        Returns:
            returns table of bands.
        """
        times = dates_times.get_time_string()
        r = str(random.randint(0, 1000))
        uid = '{}_{}'.format(times, r)
        return uid


def _create_table_of_bands(
        path, band_number=None, raster_band=None, name=None,
        multiplicative_factor=None, additive_factor=None, date=None,
        root_directory=None, wavelength=None, wavelength_unit=None
) -> np.array:
    """Creates table of bands

    This method creates a BandSet defined by input files.
    Raster properties are derived from files to populate 
    the table of bands.

    Args:
        name: optional BandSet name string
        path: file path
        band_number: sets the position of the band in BandSet order.
        raster_band: number of band for multiband rasters.
        name: name of raster used for band.
        date: single date (as YYYY-MM-DD).
        multiplicative_factor: multiplicative factor.
        additive_factor: additive factor.
        wavelength: center wavelength.
        wavelength_unit: wavelength unit.
        root_directory: root directory for relative path.
                    
    Returns:
        returns table of bands.
    """  # noqa: E501
    cfg.logger.log.debug('start')
    try:
        absolute_path = files_directories.relative_to_absolute_path(
            path, root_directory
        )
    except Exception as err:
        cfg.logger.log.error(str(err))
        if files_directories.is_file(path):
            absolute_path = path
        else:
            cfg.logger.log.error('path: %s' % path)
            messages.error('path: %s' % path)
            return None
    info = raster_vector.raster_info(absolute_path)
    if info is False:
        cfg.logger.log.error('path: %s' % path)
        messages.error('path: %s' % path)
        return None
    else:
        (gt, crs, crs_unit, xy_count, nd, number_of_bands, block_size,
         scale_offset, data_type) = info
    x_size = abs(gt[1])
    y_size = abs(gt[5])
    top = gt[3]
    left = gt[0]
    bottom = gt[3] + gt[5] * xy_count[1] + gt[4] * xy_count[0]
    right = gt[0] + gt[1] * xy_count[0] + gt[2] * xy_count[1]
    table = tm.create_band_table(
        band_number=band_number, raster_band=raster_band, path=path,
        absolute_path=absolute_path, name=name, wavelength=wavelength,
        wavelength_unit=wavelength_unit, additive_factor=additive_factor,
        multiplicative_factor=multiplicative_factor, date=date, x_size=x_size,
        y_size=y_size, top=top, left=left, bottom=bottom, right=right,
        x_count=xy_count[0], y_count=xy_count[1], nodata=nd,
        data_type=data_type, crs=crs, number_of_bands=number_of_bands,
        x_block_size=block_size[0], y_block_size=block_size[1],
        scale=scale_offset[0], offset=scale_offset[1]
    )
    cfg.logger.log.debug('end')
    return table


def _raster_to_band_names(path: str, raster_band=1) -> list:
    """Creates band name list from raster name

    Creates band name list from raster name

    Args:
        path: file path.
        raster_band: raster band number.

    Returns:
        returns list of band names.
    """
    cfg.logger.log.debug('start')
    raster_name = files_directories.file_name(path, False)
    name_list = []
    if raster_band == 1:
        name_list.append(raster_name)
    else:
        for i in range(raster_band):
            name_list.append(
                '%s%s%s' % (raster_name, cfg.band_name_suf, str(i))
            )
    cfg.logger.log.debug('end; name_list: %s' % str(name_list))
    return name_list



from unittest import TestCase

import remotior_sensus
from remotior_sensus.util import read_write_files


class TestCrossClassification(TestCase):

    def test_cross_classification(self):
        rs = remotior_sensus.Session(
            n_processes=2, available_ram=1000, log_level=10
        )
        cfg = rs.configurations
        cfg.logger.log.debug('test')
        p1 = './data/S2_2020-01-01/S2_B04.tif'
        p2 = './data/S2_2020-01-01/S2_B02.tif'
        cfg.logger.log.debug('>>> test cross_classification')
        temp0 = cfg.temp.temporary_file_path(name_suffix=cfg.tif_suffix)
        cross = rs.cross_classification(
            classification_path=p1, reference_path=p2, output_path=temp0,
            cross_matrix=True
            )
        raster, text = cross.paths
        table = read_write_files.open_text_file(text)
        table_f = read_write_files.format_csv_new_delimiter(
            table, cfg.tab_delimiter
            )
        self.assertGreater(len(table_f), 0)
        table_split = table.split(cfg.new_line)
        self.assertGreater(int(table_split[1][0]), 0)
        cfg.logger.log.debug('>>> test cross_classification accuracy matrix')
        temp1 = cfg.temp.temporary_file_path(name_suffix=cfg.tif_suffix)
        cross = rs.cross_classification(
            classification_path=p1, reference_path=p2, output_path=temp1,
            error_matrix=True
            )
        raster, text = cross.paths
        table = read_write_files.open_text_file(text)
        table_f = read_write_files.format_csv_new_delimiter(
            table, cfg.tab_delimiter
            )
        self.assertGreater(len(table_f), 0)
        table_split = table.split(cfg.new_line)
        self.assertGreater(int(table_split[1][0]), 0)
        cfg.logger.log.debug('>>> test cross_classification regression')
        temp2 = cfg.temp.temporary_file_path(name_suffix=cfg.tif_suffix)
        cross = rs.cross_classification(
            classification_path=p1, reference_path=p2, output_path=temp2,
            regression_raster=True
            )
        raster, text = cross.paths
        table = read_write_files.open_text_file(text)
        table_f = read_write_files.format_csv_new_delimiter(
            table, cfg.tab_delimiter
            )
        self.assertGreater(len(table_f), 0)
        table_split = table.split(cfg.new_line)
        self.assertGreater(int(table_split[1][0]), 0)

        # clear temporary directory
        rs.close()



from unittest import TestCase

import remotior_sensus
from remotior_sensus.util import files_directories


class TestNeighborPixels(TestCase):

    def test_neighbor_pixels(self):
        rs = remotior_sensus.Session(
            n_processes=2, available_ram=1000, log_level=10
        )
        cfg = rs.configurations
        cfg.logger.log.debug('test')
        file_list = ['./data/S2_2020-01-01/S2_B02.tif',
                     './data/S2_2020-01-01/S2_B03.tif']
        cfg.logger.log.debug('>>> test neighbor_pixels')
        neighbor = rs.band_neighbor_pixels(
            input_bands=file_list, output_path=cfg.temp.dir, size=1,
            circular_structure=True, stat_name='Mean', prefix='neighbor_'
            )
        self.assertTrue(files_directories.is_file(neighbor.paths[0]))

        # clear temporary directory
        rs.close()



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.

import os
import random

import numpy as np

from remotior_sensus.core import (
    configurations as cfg, messages, table_manager as tm
)
from remotior_sensus.core.bandset import BandSet
from remotior_sensus.core.processor_functions import spectral_signature
from remotior_sensus.util import raster_vector, dates_times, files_directories

try:
    if cfg.gdal_path is not None:
        os.add_dll_directory(cfg.gdal_path)
except Exception as error:
    cfg.logger.log.error(str(error))
try:
    from osgeo import ogr
    from osgeo import osr
except Exception as error:
    cfg.logger.log.error(str(error))
try:
    from osgeo import gdal
except Exception as error:
    cfg.logger.log.error(str(error))


class SpectralSignaturesCatalog(object):
    """A class to manage Spectral Signatures and ROIs.

    """

    def __init__(
            self, bandset: BandSet = None, catalog_table=None,
            geometry_file_path=None, macroclass_field=None, class_field=None
    ):
        # relative BandSet
        self.bandset = bandset
        # spectral signatures catalog table
        self.table = catalog_table
        # dictionary of spectral signature tables
        self.signatures = {}
        # dictionary of names of macroclasses
        self.macroclasses = {}
        # vector file linked to signature table containing ROIs
        if geometry_file_path is None:
            geometry_file_path = cfg.temp.temporary_file_path(
                name_suffix=cfg.gpkg_suffix
                )
        if not macroclass_field:
            macroclass_field = cfg.macroclass_field_name
        if not class_field:
            class_field = cfg.class_field_name
        self.macroclass_field = macroclass_field
        self.class_field = class_field
        self.geometry_file = geometry_file_path
        # create geometry vector
        if bandset:
            crs = bandset.crs
            raster_vector.create_geometry_vector(
                output_path=self.geometry_file, crs_wkt=crs,
                macroclass_field_name=self.macroclass_field,
                class_field_name=self.class_field
            )

    # add spectral signature to Spectral Signatures Catalog
    def add_spectral_signature(
            self, value_list, macroclass_id=None, class_id=None,
            macroclass_name=None, class_name=None, wavelength_list=None,
            standard_deviation_list=None, signature_id=None, selected=1,
            min_dist_thr=0, max_like_thr=0, spec_angle_thr=0
    ):
        """Adds a spectral signature.

        This method adds spectral signature to Spectral Signatures Catalog.

        Args:
            value_list:
            macroclass_id:
            class_id:
            macroclass_name:
            class_name:
            wavelength_list:
            standard_deviation_list:
            signature_id:
            selected:
            min_dist_thr:
            max_like_thr:
            spec_angle_thr:

        Returns:
            object OutputManger

        """
        if macroclass_id is None:
            macroclass_id = 1
        if class_id is None:
            class_id = 1
        # signature id
        if signature_id is None:
            signature_id = generate_signature_id()
        if wavelength_list is None:
            wavelength_list = self.bandset.get_wavelengths()
        # create signature table
        self.signatures[signature_id] = tm.create_spectral_signature_table(
            value_list=value_list, wavelength_list=wavelength_list,
            standard_deviation_list=standard_deviation_list
        )
        self.signature_to_catalog(
            signature_id=signature_id, macroclass_id=macroclass_id,
            class_id=class_id, macroclass_name=macroclass_name,
            class_name=class_name, selected=selected,
            min_dist_thr=min_dist_thr, max_like_thr=max_like_thr,
            spec_angle_thr=spec_angle_thr
        )

    # add spectral signature reference to Spectral Signatures Catalog
    def signature_to_catalog(
            self, signature_id, macroclass_id, class_id, macroclass_name=None,
            class_name=None, selected=1, min_dist_thr=0, max_like_thr=0,
            spec_angle_thr=0
    ):
        # add signature to catalog
        self.table = tm.add_spectral_signature_to_catalog_table(
            signature_id=signature_id, macroclass_id=macroclass_id,
            class_id=class_id, class_name=class_name,
            previous_catalog=self.table, selected=selected,
            min_dist_thr=min_dist_thr, max_like_thr=max_like_thr,
            spec_angle_thr=spec_angle_thr
        )
        # add or update macroclass name
        if macroclass_name is not None:
            self.macroclasses[macroclass_id] = str(macroclass_name)
        # check macroclass name
        if macroclass_id not in self.macroclasses:
            self.macroclasses[macroclass_id] = '%s%s' % (
                cfg.macroclass_default, str(len(self.macroclasses) + 1))

    # import spectral signature csv to Spectral Signatures Catalog
    def import_spectral_signature_csv(
            self, csv_path, macroclass_id=None, class_id=None,
            macroclass_name=None,
            class_name=None, separator=','
    ):
        cfg.logger.log.debug('start')
        # import csv as comma separated with fields value, wavelength,
        # standard_deviation (optional)
        if files_directories.is_file(csv_path):
            csv = tm.open_file(file_path=csv_path, separators=separator,
                               field_names=['value', 'wavelength',
                                            'standard_deviation'],
                               progress_message=False, skip_first_line=False
                               )
            bandset_wavelength = self.bandset.bands['wavelength']
            value_list = []
            wavelength_list = []
            standard_deviation_list = []
            if 'wavelength' in tm.columns(csv):
                for b in bandset_wavelength.tolist():
                    arg_min = np.abs(bandset_wavelength - b).argmin()
                    wavelength_list.append(bandset_wavelength[arg_min])
                    value_list.append(csv.value[arg_min])
                    if 'standard_deviation' in tm.columns(csv):
                        standard_deviation_list.append(
                            csv.standard_deviation[arg_min]
                        )
            if len(value_list) == 0:
                cfg.logger.log.error('file: %s' % csv_path)
                messages.error('error importing file %s' % csv_path)
                return
            if len(wavelength_list) == 0:
                wavelength_list = None
            if len(standard_deviation_list) == 0:
                standard_deviation_list = None
            self.add_spectral_signature(
                value_list=value_list, macroclass_id=macroclass_id,
                class_id=class_id,
                macroclass_name=macroclass_name, class_name=class_name,
                wavelength_list=wavelength_list,
                standard_deviation_list=standard_deviation_list
            )
            cfg.logger.log.debug('end; imported: %s' % csv_path)
        else:
            cfg.logger.log.error('error file not found: %s' % csv_path)
            messages.error('error file not found: %s' % csv_path)

    # import vector to Spectral Signatures Catalog
    def import_vector(
            self, file_path, macroclass_field, class_field,
            macroclass_name_field, class_name_field,
            calculate_signature=True
    ):
        cfg.logger.log.debug('start')
        if files_directories.is_file(file_path):
            # check geometry vector
            if not files_directories.is_file(self.geometry_file):
                if self.bandset is None:
                    raise Exception('bandset not found')
                crs = self.bandset.crs
                raster_vector.create_geometry_vector(
                    output_path=self.geometry_file, crs_wkt=crs,
                    macroclass_field_name=self.macroclass_field,
                    class_field_name=self.class_field
                )
            # get vector crs
            vector_crs = raster_vector.get_crs(file_path)
            # check crs
            catalog_sr = osr.SpatialReference()
            catalog_sr.ImportFromWkt(self.bandset.crs)
            vector_sr = osr.SpatialReference()
            vector_sr.ImportFromWkt(vector_crs)
            # required by GDAL 3 coordinate order
            try:
                catalog_sr.SetAxisMappingStrategy(
                    osr.OAMS_TRADITIONAL_GIS_ORDER
                )
                vector_sr.SetAxisMappingStrategy(
                    osr.OAMS_TRADITIONAL_GIS_ORDER
                )
            except Exception as err:
                str(err)
            if catalog_sr.IsSame(vector_sr) == 1:
                coord_transform = None
            else:
                # coordinate transformation
                coord_transform = osr.CoordinateTransformation(
                    vector_sr, catalog_sr
                )
            # open input vector
            i_vector = ogr.Open(file_path)
            i_layer = i_vector.GetLayer()
            catalog_vector = ogr.Open(self.geometry_file, 1)
            catalog_layer = catalog_vector.GetLayer()
            catalog_layer_definition = catalog_layer.GetLayerDefn()
            # import geometries
            i_feature = i_layer.GetNextFeature()
            while i_feature:
                signature_id = generate_signature_id()
                # get geometry
                geom = i_feature.GetGeometryRef()
                if coord_transform is not None:
                    # project feature
                    geom.Transform(coord_transform)
                o_feature = ogr.Feature(catalog_layer_definition)
                o_feature.SetGeometry(geom)
                macroclass_value = i_feature.GetField(macroclass_field)
                class_value = i_feature.GetField(class_field)
                macroclass_name = i_feature.GetField(macroclass_name_field)
                class_name = i_feature.GetField(class_name_field)
                o_feature.SetField(cfg.macroclass_field_name, macroclass_value)
                o_feature.SetField(cfg.class_field_name, class_value)
                o_feature.SetField(cfg.uid_field_name, signature_id)
                catalog_layer.CreateFeature(o_feature)
                if calculate_signature:
                    temp_path = cfg.temp.temporary_file_path(
                        name_suffix=cfg.gpkg_suffix
                        )
                    raster_vector.create_geometry_vector(
                        output_path=temp_path, crs_wkt=self.bandset.crs,
                        macroclass_field_name=self.macroclass_field,
                        class_field_name=self.class_field
                    )
                    temp_vector = ogr.Open(temp_path, 1)
                    temp_layer = temp_vector.GetLayer()
                    temp_layer.CreateFeature(o_feature)
                    temp_vector.Destroy()
                    value_list, standard_deviation_list = \
                        self.calculate_signature(
                            temp_path
                        )
                    self.add_spectral_signature(
                        value_list=value_list, macroclass_id=macroclass_value,
                        class_id=class_value,
                        macroclass_name=macroclass_name, class_name=class_name,
                        standard_deviation_list=standard_deviation_list,
                        signature_id=signature_id
                    )
                else:
                    self.signature_to_catalog(
                        signature_id=signature_id,
                        macroclass_id=macroclass_value, class_id=class_value,
                        macroclass_name=macroclass_name, class_name=class_name
                    )
                o_feature.Destroy()
                i_feature.Destroy()
                i_feature = i_layer.GetNextFeature()
            # close files
            i_vector.Destroy()
            catalog_vector.Destroy()
            cfg.logger.log.debug('end; imported: %s' % file_path)
        else:
            cfg.logger.log.error('error file not found: %s' % file_path)
            messages.error('error file not found: %s' % file_path)

    # import vector to Spectral Signatures Catalog
    def calculate_signature(self, roi_path, n_processes: int = None):
        if n_processes is None:
            n_processes = cfg.n_processes
        _temp_vector = ogr.Open(roi_path)
        _temp_layer = _temp_vector.GetLayer()
        min_x, max_x, min_y, max_y = _temp_layer.GetExtent()
        _temp_layer = None
        _temp_vector = None
        path_list = self.bandset.get_absolute_paths()
        virtual_path_list = []
        for p in path_list:
            temp_path = cfg.temp.temporary_file_path(
                name_suffix=cfg.tif_suffix
                )
            virtual = raster_vector.create_virtual_raster(
                input_raster_list=[p], output=temp_path,
                box_coordinate_list=[min_x, max_y, max_x, min_y]
            )
            virtual_path_list.append(virtual)
        roi_paths = [roi_path] * len(path_list)
        cfg.multiprocess.run_separated(
            raster_path_list=virtual_path_list, function=spectral_signature,
            function_argument=roi_paths, function_variable=virtual_path_list,
            n_processes=n_processes, keep_output_argument=True,
            progress_message='calculate signature', min_progress=1,
            max_progress=80
            )
        cfg.multiprocess.multiprocess_spectral_signature()
        value_list, standard_deviation_list = cfg.multiprocess.output
        return value_list, standard_deviation_list


# generate signature id
def generate_signature_id():
    times = dates_times.get_time_string()
    r = str(random.randint(0, 1000))
    uid = 's%s_%s' % (times, r)
    return uid



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.

"""
Tools to manage dates and times
"""

import datetime
from remotior_sensus.core import configurations as cfg


# get date string from name
def date_string_from_directory_name(directory_name):
    date = False
    # format YYYY-MM-DD
    try:
        datetime.datetime.strptime(directory_name[-10:], '%Y-%m-%d')
        date = directory_name[-10:]
    except Exception as err:
        str(err)
        # format YYYYMMDD
        try:
            dir_part = directory_name.split('_')
            for d_p in dir_part:
                d_p_part = d_p.lower().split('t')[0]
                try:
                    date_string = datetime.datetime.strptime(d_p_part,
                                                             '%Y%m%d')
                    d_p_part_string = date_string.strftime('%Y-%m-%d')
                    date = d_p_part_string
                    break
                except Exception as err:
                    str(err)
        except Exception as err:
            str(err)
    cfg.logger.log.debug('date: %s' % date)
    return date


# get time
def get_time_string():
    time = datetime.datetime.now().strftime('%Y%m%d_%H%M%S%f')
    return time


# get date
def get_date_string():
    time = datetime.datetime.now().strftime('%m%d')
    return time



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.

"""
Raster reclassification.

This tool allows for the reclassification of a raster based on
a reclassification table.

Typical usage example:

    >>> # import Remotior Sensus and start the session
    >>> import remotior_sensus
    >>> rs = remotior_sensus.Session()
    >>> # start the process
    >>> reclassification = rs.raster_reclassification(
    ... raster_path='file1.tif', output_path='output.tif',
    ... reclassification_table=[[1, -10], ['nan', 6000]]
    ... )
"""

from typing import Union, Optional

import numpy as np

from remotior_sensus.core import configurations as cfg, messages, \
    table_manager as tm
from remotior_sensus.core.output_manager import OutputManager
from remotior_sensus.core.processor_functions import (
    raster_unique_values_with_sum, reclassify_raster
)
from remotior_sensus.util import files_directories, raster_vector


def raster_reclassification(
        raster_path: str, output_path: Optional[str] = None,
        reclassification_table: Optional[Union[list, np.array]] = None,
        csv_path: Optional[str] = None, separator: Optional[str] = ',',
        output_data_type: Optional[str] = None,
        n_processes: Optional[int] = None,
        available_ram: Optional[int] = None,
        virtual_output: Optional[bool] = None
) -> OutputManager:
    """Performs raster reclassification.

    This tool reclassifies a raster based on a reclassification table.
    The reclassification table is defined by two columns: old values and new
    values.
    Old values define the values of the raster to be reclassified to new
    values.
    Old values can be integer numbers or conditions selecting ranges of values.

    Args:
        raster_path: path of raster used as input.
        output_path: string of output path.
        reclassification_table: table of values for reclassification or list
            of values [[old_value, new_value], ...]; if None, csv_path is used.
        csv_path: path to a csv file containing the reclassification table; 
            used if reclassification_table is None.
        separator: separator character for csv file; default is comma separated.
        output_data_type: set output data type such as 'Float32' or 'Int32'; 
            if None, input data type is used.
        n_processes: number of parallel processes.
        available_ram: number of megabytes of RAM available to processes.
        virtual_output: if True (and output_path is directory), save output
            as virtual raster of multiprocess parts.

    Returns:
        Object :func:`~remotior_sensus.core.output_manager.OutputManager` with
            - path = output path

    Examples:
        Perform the reclassification using a csv file containing the reclassification values
            >>> # import Remotior Sensus and start the session
            >>> import remotior_sensus
            >>> rs = remotior_sensus.Session()
            >>> # start the process
            >>> reclassification = rs.raster_reclassification(
            ... raster_path='file1.tif', output_path='output.tif',
            ... csv_path='file.csv'
            ... )
    """  # noqa: E501
    cfg.logger.log.info('start')
    cfg.progress.update(
        process=__name__.split('.')[-1].replace('_', ' '), message='starting',
        start=True
    )
    cfg.logger.log.debug('raster_path: %s' % str(raster_path))
    if n_processes is None:
        n_processes = cfg.n_processes
    raster_path = files_directories.input_path(raster_path)
    (gt, crs, crs_unit, xy_count, nd, number_of_bands, block_size,
     scale_offset, data_type) = raster_vector.raster_info(raster_path)
    if output_data_type is None:
        output_data_type = data_type
    # check output path
    out_path, vrt_r = files_directories.raster_output_path(
        output_path, virtual_output
    )
    files_directories.create_parent_directory(output_path)
    # reclassification table
    if reclassification_table is None:
        if csv_path is None:
            cfg.logger.log.error('unable to reclassify')
            messages.error('unable to reclassify')
            return OutputManager(check=False)
        else:
            table = _import_reclassification_table(
                csv_path=csv_path, separator=separator
                )
            if table.check:
                reclassification_table = table.extra['table']
            else:
                cfg.logger.log.error('unable to reclassify')
                messages.error('unable to reclassify')
                return OutputManager(check=False)
    # if reclassification list
    elif type(reclassification_table) is list:
        table = _list_to_reclassification_table(reclassification_table)
        if table.check:
            reclassification_table = table.extra['table']
        else:
            cfg.logger.log.error('unable to reclassify')
            messages.error('unable to reclassify')
            return OutputManager(check=False)
    # check output data type
    cfg.logger.log.debug('output_data_type: %s' % str(output_data_type))
    if (output_data_type.lower() == cfg.uint32_dt.lower()
            or output_data_type.lower() == cfg.uint16_dt.lower()):
        new_values = reclassification_table.new_value
        for i in new_values:
            try:
                if int(i) < 0:
                    output_data_type = cfg.int32_dt
                    cfg.logger.log.debug(
                        'output_data_type: %s' % str(output_data_type)
                    )
                    break
            except Exception as err:
                str(err)
    # dummy bands for memory calculation
    dummy_bands = 4
    # process calculation
    cfg.multiprocess.run(
        raster_path=raster_path, function=reclassify_raster,
        function_argument=reclassification_table,
        function_variable=cfg.variable_raster_name,
        output_raster_path=out_path, n_processes=n_processes,
        available_ram=available_ram, output_data_type=output_data_type,
        output_nodata_value=nd, compress=cfg.raster_compression,
        virtual_raster=vrt_r, progress_message='processing raster',
        dummy_bands=dummy_bands, min_progress=1, max_progress=99
    )
    cfg.progress.update(end=True)
    cfg.logger.log.info('end; band reclassification: %s' % str(out_path))
    return OutputManager(path=out_path)


def unique_values_table(
        raster_path: str, n_processes: Optional[int] = None,
        available_ram: Optional[int] = None,
        incremental: Optional[bool] = False,
        progress_message: Optional[bool] = True
):
    """Calculate unique values from raster."""
    cfg.logger.log.debug('start')
    if progress_message:
        cfg.progress.update(
            process=__name__.split('.')[-1].replace('_', ' '),
            message='starting', start=True
        )
    cfg.logger.log.debug('raster_path: %s' % str(raster_path))
    if n_processes is None:
        n_processes = cfg.n_processes
    # dummy bands for memory calculation
    dummy_bands = 2
    cfg.multiprocess.run(
        raster_path=raster_path, function=raster_unique_values_with_sum,
        keep_output_argument=True, n_processes=n_processes,
        available_ram=available_ram, dummy_bands=dummy_bands,
        progress_message='unique values', min_progress=1, max_progress=99
    )
    # calculate sum of values
    cfg.multiprocess.multiprocess_sum_array()
    values = cfg.multiprocess.output
    # add old value field
    if not incremental:
        table = tm.append_field(
            values['new_val'], cfg.old_value, values['new_val'], 'U1024'
        )
    else:
        table = tm.append_field(
            values['new_val'], cfg.old_value,
            np.arange(1, values['new_val'].shape[0] + 1, 1),
            'U1024'
        )
    # add new value field
    table = tm.append_field(table, cfg.new_value, values['new_val'], 'U1024')
    table = tm.redefine_matrix_columns(
        matrix=table, input_field_names=[cfg.old_value, cfg.new_value],
        output_field_names=[cfg.old_value, cfg.new_value]
    )
    cfg.progress.update(end=True)
    cfg.logger.log.debug('end')
    return table


def _list_to_reclassification_table(input_list):
    """# create reclassification table from list of values
    [(old value, new value),  (old value, new value)].

    configurations.variable_raster_name can be used as variable in old value
    e.g.
    raster > 0.3
    """
    # table of values
    table = None
    dtype_list = [(cfg.old_value, 'U1024'), (cfg.new_value, 'U1024')]
    # test array
    _x = np.ones(3)
    for i in range(len(input_list)):
        # check new value
        try:
            new_value = str(input_list[i][1]).lower().replace('null', 'nan')
            eval(new_value.replace('nan', 'np.nan'))
        except Exception as err:
            cfg.logger.log.error(str(err))
            messages.error(str(err))
            return OutputManager(check=False)
        # check old value
        try:
            old_value = int(input_list[i][0])
        except Exception as err:
            str(err)
            try:
                old_value = input_list[i][0].lower().replace('null', 'nan')
                test_old_value = old_value.replace(
                    cfg.variable_raster_name, '_x'
                ).replace('nan', 'np.nan')
                eval(test_old_value)
            except Exception as err:
                str(err)
                cfg.logger.log.error(
                    'unable to process value %s' % str(input_list[i][0])
                )
                messages.error(
                    'unable to process value %s' % str(input_list[i][0])
                )
                return OutputManager(check=False)
        # add to table
        if table is None:
            table = np.array(
                (str(old_value), str(new_value)), dtype=dtype_list
            )
        else:
            table = tm.append_values_to_table(
                table, [str(old_value), str(new_value)]
            )
    return OutputManager(extra={'table': table})


def _import_reclassification_table(csv_path, separator=','):
    """Imports reclassification table from csv with two columns of values
    (old value, new value)."""
    # open csv
    csv = tm.open_file(
        file_path=csv_path, separators=separator,
        field_names=[cfg.old_value, cfg.new_value], progress_message=False,
        skip_first_line=False
        )
    # table of values
    table = None
    dtype_list = [(cfg.old_value, 'U1024'), (cfg.new_value, 'U1024')]
    # test array
    _x = np.ones(3)
    for i in range(len(csv)):
        # check new value
        try:
            new_value = str(csv[cfg.new_value][i]).lower().replace(
                'null', 'nan'
            )
            eval(new_value.replace('nan', 'np.nan'))
        except Exception as err:
            cfg.logger.log.error(str(err))
            messages.error(str(err))
            return OutputManager(check=False)
        # check old value
        try:
            old_value = int(csv[cfg.old_value][i])
        except Exception as err:
            str(err)
            try:
                old_value = csv[cfg.old_value][i].lower().replace(
                    'null', 'nan'
                )
                test_old_value = old_value.replace(
                    cfg.variable_raster_name, '_x'
                )
                eval(test_old_value)
            except Exception as err:
                str(err)
                cfg.logger.log.error(
                    'unable to process value %s' % str(csv[cfg.old_value][i])
                )
                messages.error(
                    'unable to process value %s' % str(csv[cfg.old_value][i])
                )
                return OutputManager(check=False)
        # add to table
        if table is None:
            table = np.array(
                (str(old_value), str(new_value)), dtype=dtype_list
            )
        else:
            table = tm.append_values_to_table(
                table, [str(old_value), str(new_value)]
            )
    return OutputManager(extra={'table': table})



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.

from contextlib import redirect_stdout

import numpy as np

from remotior_sensus.core import configurations as cfg
from remotior_sensus.util import raster_vector

try:
    from sklearn.model_selection import cross_val_score, StratifiedKFold
except Exception as error:
    str(error)
try:
    import scipy.stats.distributions as statdistr
except Exception as error:
    str(error)
try:
    from scipy import signal
except Exception as error:
    str(error)
try:
    from scipy.ndimage.filters import maximum_filter
    from scipy.ndimage.filters import minimum_filter
    from scipy.ndimage.filters import percentile_filter
    from scipy.ndimage.filters import generic_filter
    from scipy.ndimage.filters import median_filter
except Exception as error:
    str(error)

try:
    import torch
    from remotior_sensus.util.pytorch_tools import train_pytorch_model
except Exception as error:
    str(error)


# band calculation
def band_calculation(*argv):
    # expose numpy functions
    log = np.log
    _log = log
    log10 = np.log10
    _log10 = log10
    sqrt = np.sqrt
    _sqrt = sqrt
    cos = np.cos
    _cos = cos
    arccos = np.arccos
    _arccos = arccos
    sin = np.sin
    _sin = sin
    arcsin = np.arcsin
    _arcsin = arcsin
    tan = np.tan
    _tan = tan
    arctan = np.arctan
    _arctan = arctan
    exp = np.exp
    _exp = exp
    min = np.nanmin
    _min = min
    max = np.nanmax
    _max = max
    sum = np.nansum
    _sum = sum
    percentile = np.nanpercentile
    _percentile = percentile
    median = np.nanmedian
    _median = median
    mean = np.nanmean
    _mean = mean
    std = np.nanstd
    _std = std
    where = np.where
    _where = where
    nan = np.nan
    _nan = nan
    # array variable name as defined in cfg.array_function_placeholder
    _array_function_placeholder = argv[1]
    nodata_mask = argv[2]
    function_argument = argv[7]
    cfg.logger.log.debug('function_argument: %s' % str(function_argument))
    if (_array_function_placeholder.dtype == np.float32
            or _array_function_placeholder.dtype == np.float64):
        cfg.logger.log.debug(
            '_array_function_placeholder.shape: %s; '
            '_array_function_placeholder.dtype: %s; '
            '_array_function_placeholder.nbytes: %s'
            % (
                str(_array_function_placeholder.shape),
                str(_array_function_placeholder.dtype),
                str(_array_function_placeholder.nbytes)
               )
        )
        _array_function_placeholder = ArrayLike(_array_function_placeholder)
    # perform operation
    try:
        _o = eval(function_argument)
    except Exception as err:
        cfg.logger.log.error(str(err))
        return False
    # if not array
    if not isinstance(_o, np.ndarray):
        cfg.logger.log.error('not array')
        return False
    # check nodata
    cfg.logger.log.debug(
        '_o.shape: %s; nodata_mask.shape: %s; _o.nbytes: %s; _o.dtype: %s'
        % (
            str(_o.shape), str(nodata_mask.shape), str(_o.nbytes),
            str(_o.dtype)
        )
    )
    if nodata_mask is not None:
        np.copyto(
            _o, nodata_mask.reshape(_o.shape),
            where=nodata_mask[::, ::].reshape(_o.shape) != 0
        )
    return [_o, None]


# classification maximum likelihood
def classification_maximum_likelihood(*argv):
    scale = argv[0][0]
    offset = argv[0][1]
    output_no_data = argv[0][2]
    _array_function_placeholder = argv[1]
    nodata_mask = argv[2]
    x = argv[4]
    y = argv[5]
    # function argument
    signatures = argv[7][cfg.spectral_signatures_framework]
    # get selected signatures
    signatures_table = signatures.table[signatures.table.selected == 1]
    covariance_matrices = argv[7][cfg.covariance_matrices_framework]
    normalization = argv[7][cfg.normalization_values_framework]
    function_variable_list = argv[8]
    macroclass = function_variable_list[0]
    threshold = function_variable_list[1]
    ro_x, ro_y = argv[10]
    output_signature_raster = argv[11]
    out_class = argv[12]
    out_alg = argv[13]
    previous_array = None
    classification_array = None
    if normalization[0] is not None:
        for n in range(0, _array_function_placeholder.shape[2]):
            _array_function_placeholder[::, ::, n] = eval(normalization[0][n])
    for s in signatures_table.signature_id.tolist():
        if macroclass:
            class_id = signatures_table[
                signatures_table.signature_id == s].macroclass_id
        else:
            class_id = signatures_table[
                signatures_table.signature_id == s].class_id
        values = signatures.signatures[s].value
        cfg.logger.log.debug('signature: %s; values: %s' % (s, str(values)))
        try:
            cov_matrix = covariance_matrices[s]
            # natural logarithm of the determinant of covariance matrix
            (sign, log_det) = np.linalg.slogdet(cov_matrix)
            inverse_cov_matrix = np.linalg.inv(cov_matrix)
            d = _array_function_placeholder - values
            distance_array = - log_det - (
                    np.dot(d, inverse_cov_matrix) * d).sum(axis=2)
            if threshold:
                class_threshold = signatures_table[
                    signatures_table.signature_id == s].min_dist_thr
                p = class_threshold / 100
                chi = statdistr.chi2.isf(p, cov_matrix.shape[0])
                chi_threshold = -2 * chi - log_det
                distance_array[::, ::][
                    distance_array < chi_threshold] = cfg.nodata_val
            if previous_array is None:
                previous_array = distance_array
                classification_array = np.ones(
                    (_array_function_placeholder.shape[0],
                     _array_function_placeholder.shape[1])
                ) * class_id
            else:
                maximum_array = np.maximum(distance_array, previous_array)
                classification_array[
                    maximum_array != previous_array] = class_id
                previous_array = maximum_array
            if len(output_signature_raster) > 0:
                distance_array[::, ::][
                    distance_array == cfg.nodata_val] = output_no_data
                distance_array[::, ::][
                    nodata_mask == output_no_data] = output_no_data
                write_sig = raster_vector.write_raster(
                    output_signature_raster[s], x - ro_x, y - ro_y,
                    distance_array, output_no_data, scale, offset
                )
                cfg.logger.log.debug('write_sig: %s' % str(write_sig))
        except Exception as err:
            cfg.logger.log.error(str(err))
    if classification_array is not None:
        # write classification
        classification_array[::, ::][
            classification_array == cfg.nodata_val] = output_no_data
        classification_array[::, ::][
            nodata_mask == output_no_data] = output_no_data
        write_class = raster_vector.write_raster(
            out_class, x - ro_x, y - ro_y, classification_array, output_no_data,
            scale, offset
        )
        cfg.logger.log.debug('write_class: %s' % str(write_class))
        # write the algorithm raster
        if out_alg is not None:
            previous_array[::, ::][
                classification_array == cfg.nodata_val] = output_no_data
            previous_array[::, ::][nodata_mask == output_no_data] = output_no_data
            write_alg = raster_vector.write_raster(
                out_alg, x - ro_x, y - ro_y, previous_array, output_no_data, scale,
                offset
            )
            cfg.logger.log.debug('write_alg: %s' % str(write_alg))
        cfg.logger.log.debug(
            'classification_array.shape: %s' % str(classification_array.shape)
        )
        return [True, out_class]
    else:
        return [False, out_class]


# classification minimum distance
def classification_minimum_distance(*argv):
    scale = argv[0][0]
    offset = argv[0][1]
    output_no_data = argv[0][2]
    _array_function_placeholder = argv[1]
    nodata_mask = argv[2]
    x = argv[4]
    y = argv[5]
    signatures = argv[7][cfg.spectral_signatures_framework]
    # get selected signatures
    signatures_table = signatures.table[signatures.table.selected == 1]
    normalization = argv[7][cfg.normalization_values_framework]
    function_variable_list = argv[8]
    macroclass = function_variable_list[0]
    threshold = function_variable_list[1]
    ro_x, ro_y = argv[10]
    output_signature_raster = argv[11]
    out_class = argv[12]
    out_alg = argv[13]
    previous_array = None
    classification_array = None
    if normalization[0] is not None:
        for n in range(0, _array_function_placeholder.shape[2]):
            _array_function_placeholder[::, ::, n] = eval(normalization[0][n])
    for s in signatures_table.signature_id.tolist():
        if macroclass:
            class_id = signatures_table[
                signatures_table.signature_id == s].macroclass_id
        else:
            class_id = signatures_table[
                signatures_table.signature_id == s].class_id
        values = signatures.signatures[s].value
        # euclidean distance
        distance_array = np.sqrt(
            ((_array_function_placeholder - values) ** 2).sum(axis=2)
        )
        if threshold:
            class_threshold = signatures_table[
                signatures_table.signature_id == s].min_dist_thr
            distance_array[::, ::][
                distance_array < class_threshold] = cfg.nodata_val_Int32
        if previous_array is None:
            previous_array = distance_array
            classification_array = np.ones(
                (_array_function_placeholder.shape[0],
                 _array_function_placeholder.shape[1])
            ) * class_id
        else:
            minimum_array = np.minimum(distance_array, previous_array)
            classification_array[minimum_array != previous_array] = class_id
            previous_array = minimum_array
        if len(output_signature_raster) > 0:
            distance_array[::, ::][
                distance_array == cfg.nodata_val_Int32] = output_no_data
            distance_array[::, ::][
                nodata_mask == output_no_data] = output_no_data
            write_sig = raster_vector.write_raster(
                output_signature_raster[s], x - ro_x, y - ro_y, distance_array,
                output_no_data, scale, offset
            )
            cfg.logger.log.debug('write_sig: %s' % str(write_sig))
    # write classification
    classification_array[::, ::][
        classification_array == cfg.nodata_val_Int32] = output_no_data
    classification_array[::, ::][
        nodata_mask == output_no_data] = output_no_data
    write_class = raster_vector.write_raster(
        out_class, x - ro_x, y - ro_y, classification_array, output_no_data,
        scale, offset
    )
    cfg.logger.log.debug('write_class: %s' % str(write_class))
    # write the algorithm raster
    if out_alg is not None:
        previous_array[::, ::][
            classification_array == cfg.nodata_val_Int32] = output_no_data
        previous_array[::, ::][nodata_mask == output_no_data] = output_no_data
        write_alg = raster_vector.write_raster(
            out_alg, x - ro_x, y - ro_y, previous_array, output_no_data, scale,
            offset
        )
        cfg.logger.log.debug('write_alg: %s' % str(write_alg))
    cfg.logger.log.debug(
        'classification_array.shape: %s' % str(classification_array.shape)
    )
    return [True, out_class]


# classification spectral angle mapping
def classification_spectral_angle_mapping(*argv):
    scale = argv[0][0]
    offset = argv[0][1]
    output_no_data = argv[0][2]
    _array_function_placeholder = argv[1]
    nodata_mask = argv[2]
    x = argv[4]
    y = argv[5]
    signatures = argv[7][cfg.spectral_signatures_framework]
    # get selected signatures
    signatures_table = signatures.table[signatures.table.selected == 1]
    normalization = argv[7][cfg.normalization_values_framework]
    function_variable_list = argv[8]
    macroclass = function_variable_list[0]
    threshold = function_variable_list[1]
    ro_x, ro_y = argv[10]
    output_signature_raster = argv[11]
    out_class = argv[12]
    out_alg = argv[13]
    previous_array = None
    classification_array = None
    if normalization[0] is not None:
        for n in range(0, _array_function_placeholder.shape[2]):
            _array_function_placeholder[::, ::, n] = eval(normalization[0][n])
    for s in signatures_table.signature_id.tolist():
        if macroclass:
            class_id = signatures_table[
                signatures_table.signature_id == s].macroclass_id
        else:
            class_id = signatures_table[
                signatures_table.signature_id == s].class_id
        values = signatures.signatures[s].value
        # spectral angle
        distance_array = np.arccos(
            (_array_function_placeholder * values).sum(axis=2) / np.sqrt(
                (_array_function_placeholder ** 2).sum(axis=2) * (
                        values ** 2).sum()
            )
        ) * 180 / np.pi
        if threshold:
            class_threshold = signatures_table[
                signatures_table.signature_id == s].min_dist_thr
            distance_array[::, ::][
                distance_array < class_threshold] = cfg.nodata_val_Int32
        if previous_array is None:
            previous_array = distance_array
            classification_array = np.ones(
                (_array_function_placeholder.shape[0],
                 _array_function_placeholder.shape[1])
            ) * class_id
        else:
            minimum_array = np.minimum(distance_array, previous_array)
            classification_array[minimum_array != previous_array] = class_id
            previous_array = minimum_array
        if len(output_signature_raster) > 0:
            distance_array[::, ::][
                distance_array == cfg.nodata_val_Int32] = output_no_data
            distance_array[::, ::][
                nodata_mask == output_no_data] = output_no_data
            write_sig = raster_vector.write_raster(
                output_signature_raster[s], x - ro_x, y - ro_y, distance_array,
                output_no_data, scale, offset
            )
            cfg.logger.log.debug('write_sig: %s' % str(write_sig))
    # write classification
    classification_array[::, ::][
        classification_array == cfg.nodata_val_Int32] = output_no_data
    classification_array[::, ::][
        nodata_mask == output_no_data] = output_no_data
    write_class = raster_vector.write_raster(
        out_class, x - ro_x, y - ro_y, classification_array, output_no_data,
        scale, offset
    )
    cfg.logger.log.debug('write_class: %s' % str(write_class))
    # write the algorithm raster
    if out_alg is not None:
        previous_array[::, ::][
            classification_array == cfg.nodata_val_Int32] = output_no_data
        previous_array[::, ::][nodata_mask == output_no_data] = output_no_data
        write_alg = raster_vector.write_raster(
            out_alg, x - ro_x, y - ro_y, previous_array, output_no_data, scale,
            offset
        )
        cfg.logger.log.debug('write_alg: %s' % str(write_alg))
    cfg.logger.log.debug(
        'classification_array.shape: %s' % str(classification_array.shape)
    )
    return [True, out_class]


# classification through scikit-learn model
def classification_scikit(*argv):
    scale = argv[0][0]
    offset = argv[0][1]
    output_no_data = argv[0][2]
    _array_function_placeholder = argv[1]
    nodata_mask = argv[2]
    x = argv[4]
    y = argv[5]
    classifier = argv[7][cfg.model_classifier_framework]
    cfg.logger.log.debug('classifier: %s' % str(classifier.classes_))
    normalization = argv[7][cfg.normalization_values_framework]
    function_variable_list = argv[8]
    x_min_piece, y_min_piece = argv[10]
    threshold = function_variable_list[1]
    out_class = argv[12]
    out_alg = argv[13]
    x_array = None
    if normalization[0] is not None:
        for n in range(0, _array_function_placeholder.shape[2]):
            _array_function_placeholder[::, ::, n] = eval(normalization[0][n])
    for n in range(_array_function_placeholder.shape[2]):
        if x_array is None:
            x_array = _array_function_placeholder[::, ::, n].ravel()
        else:
            x_array = np.vstack(
                [x_array, _array_function_placeholder[::, ::, n].ravel()]
            )
    # replace nan
    x_array[np.isnan(x_array)] = cfg.nodata_val
    # prediction
    if not threshold and out_alg is None:
        _prediction = classifier.predict(x_array.T)
        classification_array = _prediction.reshape(
            _array_function_placeholder.shape[0],
            _array_function_placeholder.shape[1]
        )
        _prediction = None
        # write classification
        classification_array[::, ::][
            nodata_mask == output_no_data] = output_no_data
        try:
            write_class = raster_vector.write_raster(
                out_class, x - x_min_piece, y - y_min_piece,
                classification_array, output_no_data, scale, offset
            )
            cfg.logger.log.debug(
                'write_class: %s; classification_array.shape: %s'
                % (str(write_class), str(classification_array.shape))
            )
        except Exception as err:
            cfg.logger.log.debug(
                'classification_array.shape: %s; x: %s; y: %s: '
                % (str(classification_array.shape), str(x), str(y))
            )
            cfg.logger.log.error(str(err))
    # write the probability raster
    elif out_alg is not None:
        try:
            _prediction_proba = classifier.predict_proba(x_array.T)
            prediction_proba_array = np.max(_prediction_proba, axis=1).reshape(
                _array_function_placeholder.shape[0],
                _array_function_placeholder.shape[1]
            )
            classification_argmax = np.argmax(
                _prediction_proba, axis=1
            ).reshape(
                _array_function_placeholder.shape[0],
                _array_function_placeholder.shape[1]
            )
            _prediction_proba = None
            classes = classifier.classes_
            classification_array = np.zeros_like(classification_argmax)
            for c in range(0, len(classes)):
                classification_array = np.where(
                    classification_argmax == c, classes[c],
                    classification_array
                )
            if threshold is not False:
                classification_array[::, ::][
                    prediction_proba_array < threshold] = output_no_data
            prediction_proba_array[::, ::][
                nodata_mask == output_no_data] = output_no_data
            write_alg = raster_vector.write_raster(
                out_alg, x - x_min_piece, y - y_min_piece,
                prediction_proba_array, output_no_data, scale, offset
            )
            cfg.logger.log.debug('write_alg: %s' % str(write_alg))
        except Exception as err:
            cfg.logger.log.error(str(err))
            _prediction = classifier.predict(x_array.T)
            classification_array = _prediction.reshape(
                _array_function_placeholder.shape[0],
                _array_function_placeholder.shape[1]
            )
            _prediction = None
            # write classification
            classification_array[::, ::][
                nodata_mask == output_no_data] = output_no_data
        # write classification
        classification_array[::, ::][
            nodata_mask == output_no_data] = output_no_data
        write_class = raster_vector.write_raster(
            out_class, x - x_min_piece, y - y_min_piece, classification_array,
            output_no_data, scale, offset
        )
        cfg.logger.log.debug(
            'write_class: %s; classification_array.shape: %s'
            % (str(write_class), str(classification_array.shape))
        )
    return [True, out_class]


# classification through pytorch model
def classification_pytorch(*argv):
    scale = argv[0][0]
    offset = argv[0][1]
    output_no_data = argv[0][2]
    _array_function_placeholder = argv[1]
    nodata_mask = argv[2]
    x = argv[4]
    y = argv[5]
    classifier = argv[7][cfg.model_classifier_framework]
    normalization = argv[7][cfg.normalization_values_framework]
    function_variable_list = argv[8]
    x_min_piece, y_min_piece = argv[10]
    threshold = function_variable_list[1]
    out_class = argv[12]
    out_alg = argv[13]
    x_array = None
    if normalization[0] is not None:
        for n in range(0, _array_function_placeholder.shape[2]):
            _array_function_placeholder[::, ::, n] = eval(normalization[0][n])
    for n in range(_array_function_placeholder.shape[2]):
        if x_array is None:
            x_array = _array_function_placeholder[::, ::, n].ravel()
        else:
            x_array = np.vstack(
                [x_array, _array_function_placeholder[::, ::, n].ravel()]
            )
    # replace nan
    x_array[np.isnan(x_array)] = cfg.nodata_val
    # prediction
    torch.set_num_threads(1)
    classifier.eval()
    data_type = eval('torch.%s' % str(x_array.dtype))
    x_array = torch.tensor(x_array.T, dtype=data_type)
    if not threshold and out_alg is None:
        _prediction = classifier(x_array).argmax(1).numpy()
        classification_array = _prediction.reshape(
            _array_function_placeholder.shape[0],
            _array_function_placeholder.shape[1]
        )
        _prediction = None
        # write classification
        classification_array[::, ::][
            nodata_mask == output_no_data] = output_no_data
        try:
            write_class = raster_vector.write_raster(
                out_class, x - x_min_piece, y - y_min_piece,
                classification_array, output_no_data, scale, offset
            )
            cfg.logger.log.debug('write_class: %s' % str(write_class))
            cfg.logger.log.debug(
                'classification_array.shape: %s' % str(
                    classification_array.shape
                )
            )
        except Exception as err:
            cfg.logger.log.debug(
                'classification_array.shape: %s; x: %s; y: %s'
                % (str(classification_array.shape), str(x), str(y))
            )
            cfg.logger.log.error(str(err))
    # write the probability raster
    elif out_alg is not None:
        try:
            _softmax = torch.softmax(classifier(x_array), dim=1)
            prediction_proba_array, classification_array = _softmax.topk(
                1, dim=1
            )
            _softmax = None
            prediction_proba_array = prediction_proba_array.detach().numpy(

            ).reshape(
                _array_function_placeholder.shape[0],
                _array_function_placeholder.shape[1]
            )
            classification_array = classification_array.detach().numpy(
            ).reshape(
                _array_function_placeholder.shape[0],
                _array_function_placeholder.shape[1]
            )
            if threshold is not False:
                classification_array[::, ::][
                    prediction_proba_array < threshold] = output_no_data
            prediction_proba_array[::, ::][
                nodata_mask == output_no_data] = output_no_data
            write_alg = raster_vector.write_raster(
                out_alg, x - x_min_piece, y - y_min_piece,
                prediction_proba_array, output_no_data, scale, offset
            )
            cfg.logger.log.debug('write_alg: %s' % str(write_alg))
        except Exception as err:
            cfg.logger.log.error(str(err))
            prediction = classifier(x_array).argmax(1).numpy()
            classification_array = prediction.reshape(
                _array_function_placeholder.shape[0],
                _array_function_placeholder.shape[1]
            )
            # write classification
            classification_array[::, ::][
                nodata_mask == output_no_data] = output_no_data
        # write classification
        classification_array[::, ::][
            nodata_mask == output_no_data] = output_no_data
        write_class = raster_vector.write_raster(
            out_class, x - x_min_piece, y - y_min_piece, classification_array,
            output_no_data, scale, offset
        )
        cfg.logger.log.debug(
            'write_class: %s; classification_array.shape: %s'
            % (str(write_class), str(classification_array.shape))
        )
    return [True, out_class]


# calculate PCA
def calculate_pca(*argv):
    # array variable name as defined in cfg.array_function_placeholder
    _array_function_placeholder = argv[1]
    nodata_mask = argv[2]
    # principal component vector
    function_argument = argv[7]
    # band mean vector
    function_variable = argv[8]
    cfg.logger.log.debug(
        'function_argument: %s; function_variable: %s'
        % (str(function_argument), str(function_variable))
    )
    # perform calculation
    _o = (function_argument * (
            _array_function_placeholder - function_variable
    )).sum(axis=2, dtype=np.float32)
    # check nodata
    cfg.logger.log.debug(
        '_o: %s; nodata_mask: %s' % (str(_o.shape), str(nodata_mask.shape))
    )
    if nodata_mask is not None:
        np.copyto(
            _o, nodata_mask.reshape(_o.shape),
            where=nodata_mask[::, ::].reshape(_o.shape) != 0
        )
    return [_o, None]


# reclassify raster
def reclassify_raster(*argv):
    output_no_data = argv[0][2]
    raster_array_band = argv[1]
    nodata_mask = argv[2]
    # conditions
    function_argument = argv[7]
    # variable raster name
    function_variable = argv[8]
    cfg.logger.log.debug('start')
    _o = None
    replace_nodata = True
    try:
        old = function_argument.old_value
        new = function_argument.old_value
        _raster = np.nan_to_num(raster_array_band[:, :, 0])
        # if all integer values
        if np.all(_raster.astype(int) == _raster) and np.all(
                old.astype(int) == old
        ):
            # create empty reclass array of length equal to maximum value
            reclass = np.zeros(
                max(
                    old.astype(int).max(), raster_array_band.astype(int).max()
                ) + 1
            ) * np.nan
            # fill array with new values at index corresponding to old value
            reclass[old.astype(int)] = new
            # perform reclassification
            _o = reclass[_raster.astype(int)]
        else:
            # raise exception to try expressions
            raise Exception
    except Exception as err:
        str(err)
        _raster = None
        # raster array
        _o = np.copy(raster_array_band[:, :, 0])
        _x = raster_array_band[:, :, 0]
        for i in range(function_argument.shape[0]):
            cfg.logger.log.debug(str(function_argument[i]))
            # if reclassify from nodata to new value
            if 'nan' in function_argument[i][cfg.old_value]:
                try:
                    # replace nodata
                    _o[::, ::][nodata_mask == output_no_data] = int(
                        float(function_argument[i][cfg.new_value])
                    )
                    replace_nodata = False
                except Exception as err:
                    str(err)
            else:
                # create condition for single values
                try:
                    _o[_x == int(function_argument[i][cfg.old_value])] = int(
                        float(
                            function_argument[i][
                                cfg.new_value].lower().replace(
                                'np.nan', str(output_no_data)
                            ).replace('nan', str(output_no_data))
                        )
                    )
                except Exception as err:
                    str(err)
                    # execute conditional expression
                    try:
                        exp = ('_o[%s] = %s' % (
                            function_argument[i][cfg.old_value].replace(
                                function_variable, '_x'
                            ),
                            str(
                                int(
                                    float(
                                        function_argument[i][
                                            cfg.new_value].lower().replace(
                                            'np.nan', str(output_no_data)
                                        ).replace('nan', str(output_no_data))
                                    )
                                )
                            )
                        )
                               )
                        exec(exp)
                    except Exception as err:
                        cfg.logger.log.error(
                            '%s; function_argument[i]: %s' % (
                                str(err), str(function_argument[i]))
                        )
                        break
    if replace_nodata:
        try:
            # replace nodata
            _o[::, ::][nodata_mask == output_no_data] = output_no_data
        except Exception as err:
            str(err)
    cfg.logger.log.debug('end')
    return [_o, None]


# calculate bands covariance
def bands_covariance(*argv):
    cfg.logger.log.debug('start')
    output_no_data = argv[0][2]
    raster_array_band = argv[1]
    nodata_mask = argv[2]
    band_number = argv[7]
    band_dict = argv[8]
    covariance_dictionary = {}
    # iterate bands
    for _x in band_number:
        # calculate covariance SUM((x - Mean_x) * (y - Mean_y))
        for _y in band_number:
            # mask nodata
            if _x == _y:
                x = raster_array_band[::, ::, _x][
                    nodata_mask != output_no_data].ravel()
                y = x
            else:
                x = raster_array_band[::, ::, _x][
                    nodata_mask != output_no_data].ravel()
                y = raster_array_band[::, ::, _y][
                    nodata_mask != output_no_data].ravel()
            # covariance
            cov = ((x - band_dict['mean_%s' % _x]) * (
                    y - band_dict['mean_%s' % _y])).sum()
            covariance_dictionary['cov_%s-%s' % (_x, _y)] = cov
    cfg.logger.log.debug('end')
    return [None, covariance_dictionary]


# calculate raster pixel count
def raster_pixel_count(*argv):
    cfg.logger.log.debug('start')
    output_no_data = argv[0][2]
    raster_array_band = argv[1]
    nodata_mask = argv[2]
    band_number = argv[7]
    band = raster_array_band[nodata_mask != output_no_data].ravel()
    count = band.shape[0]
    band_sum = np.nansum(band)
    raster_dictionary = {
        'count_%s' % str(band_number): count,
        'sum_%s' % str(band_number): band_sum
    }
    cfg.logger.log.debug('end')
    return [None, raster_dictionary]


# calculate raster unique values with sum
def raster_unique_values_with_sum(*argv):
    cfg.logger.log.debug('start')
    output_no_data = argv[0][2]
    raster_array_band = argv[1]
    try:
        nodata_mask = argv[2]
        b = raster_array_band[nodata_mask != output_no_data]
        stats = np.array(np.unique(b[~np.isnan(b)], return_counts=True))
    except Exception as err:
        stats = np.array(
            np.unique(
                raster_array_band[~np.isnan(raster_array_band)],
                return_counts=True
            )
        )
        cfg.logger.log.error(str(err))
    cfg.logger.log.debug('end')
    return [None, stats]


# calculate raster unique values of combinations
def raster_unique_values(*argv):
    cfg.logger.log.debug('start')
    output_no_data = argv[0][2]
    raster_array_band = argv[1]
    nodata_mask = argv[2]
    # stack arrays
    try:
        arr = raster_array_band[:, :, 0][nodata_mask != output_no_data].ravel()
        for i in range(1, raster_array_band.shape[2]):
            arr = np.vstack(
                (arr, raster_array_band[:, :, i][
                    nodata_mask != output_no_data].ravel())
            )
    except Exception as err:
        str(err)
        arr = raster_array_band[:, :, 0].ravel()
        for i in range(1, raster_array_band.shape[2]):
            arr = np.vstack((arr, raster_array_band[:, :, i].ravel()))
    arr = arr.T
    arr = arr[~np.isnan(arr).any(axis=1)]
    # adapted from Jaime answer at
    # https://stackoverflow.com/questions/16970982/find-unique-rows-in-numpy-array
    b = arr.view(np.dtype((np.void, arr.dtype.itemsize * arr.shape[1])))
    ff, index_a = np.unique(b, return_index=True, return_counts=False)
    cfg.logger.log.debug('end')
    return [None, arr[index_a]]


# calculate raster dilation
def raster_dilation(*argv):
    cfg.logger.log.debug('start')
    output_no_data = argv[0][2]
    raster_array_band = argv[1]
    nodata_mask = argv[2]
    structure = argv[7]
    function_variable_list = argv[8]
    a = np.nan_to_num(raster_array_band[:, :, 0])
    # value dictionary
    val_dict = {}
    # calculate
    for i in function_variable_list:
        try:
            val_dict['arr_%s' % str(i)] = (signal.oaconvolve(
                a == i, structure, mode='same'
            ) > 0.999)
        except Exception as err:
            str(err)
            # if scipy version < 1.4
            val_dict['arr_%s' % str(i)] = (
                    signal.fftconvolve(a == i, structure, mode='same') > 0.999)
    # core
    var_array = np.array(function_variable_list)
    core = ~np.isin(a, var_array)
    # dilation
    o = np.array(a, copy=True)
    try:
        for v in function_variable_list:
            o[(core * val_dict['arr_%s' % str(v)]) > 0] = v
    except Exception as err:
        cfg.logger.log.error(str(err))
    o[::, ::][np.isnan(raster_array_band[:, :, 0])] = np.nan
    o[::, ::][nodata_mask == output_no_data] = np.nan
    cfg.logger.log.debug('end')
    return [o, None]


# calculate raster erosion
def raster_erosion(*argv):
    cfg.logger.log.debug('start')
    output_no_data = argv[0][2]
    raster_array_band = argv[1]
    nodata_mask = argv[2]
    structure = argv[7]
    function_variable_list = argv[8]
    a = np.array(raster_array_band[::, ::, 0], copy=True)
    np.copyto(a, cfg.nodata_val, where=np.isnan(raster_array_band[:, :, 0]))
    np.copyto(a, cfg.nodata_val, where=nodata_mask == output_no_data)
    # unique value list
    unique_val = np.unique(a)
    unique_val_list = list(unique_val.astype(int))
    try:
        unique_val_list.remove(cfg.nodata_val)
    except Exception as err:
        str(err)
    # iteration of erosion size
    for _s in range(function_variable_list[0]):
        # empty array
        erosion = np.zeros(a.shape)
        # structure core pixels
        try:
            sum_structure = signal.oaconvolve(
                np.ones(a.shape), structure, mode='same'
            )
        except Exception as err:
            str(err)
            # if scipy version < 1.4
            sum_structure = signal.fftconvolve(
                np.ones(a.shape), structure, mode='same'
            )
        # value that fills erosion
        fill_value = np.ones(a.shape) * float(cfg.nodata_val)
        # maximum of all value convolution
        max_sum_unique = np.zeros(a.shape)
        # iteration of erosion values
        for i in unique_val_list:
            # frequency of values
            try:
                sum_unique = signal.oaconvolve(a == i, structure, mode='same')
            except Exception as err:
                str(err)
                # if scipy version < 1.4
                sum_unique = signal.fftconvolve(a == i, structure, mode='same')
            # expand
            if i not in function_variable_list[1]:
                # fill with most frequent value
                fill_value[sum_unique > max_sum_unique] = float(i)
                # maximum of all value convolution
                max_sum_unique[sum_unique > max_sum_unique] = sum_unique[
                    sum_unique > max_sum_unique]
            # erode
            else:
                # erosion values
                erosion[((sum_structure - sum_unique) > 0.01) & (a == i)] = 1
        np.copyto(a, fill_value, where=erosion == 1)
    a[::, ::][a == cfg.nodata_val] = np.nan
    a[::, ::][np.isnan(raster_array_band[:, :, 0])] = np.nan
    a[::, ::][nodata_mask == output_no_data] = np.nan
    cfg.logger.log.debug('end')
    return [a, None]


# calculate raster neighbor
def raster_neighbor(*argv):
    cfg.logger.log.debug('start')
    output_no_data = argv[0][2]
    raster_array_band = argv[1]
    nodata_mask = argv[2]
    structure = argv[7]
    function_variable_list = argv[8]
    cfg.logger.log.debug(
        'structure.shape: %s; function_variable_list[0]: %s; '
        'raster_array_band[0, 0 , 0]: %s'
        % (str(structure.shape), str(function_variable_list[0]),
           str(raster_array_band[0, 0, 0]))
    )
    o = None
    # calculate
    if 'nansum' in function_variable_list[0]:
        try:
            o = np.round(
                signal.oaconvolve(
                    np.nan_to_num(raster_array_band[:, :, 0]), structure,
                    mode='same'
                ), 6
            )
        except Exception as err:
            cfg.logger.log.error(str(err))
            # if scipy version < 1.4
            o = np.round(
                signal.fftconvolve(
                    np.nan_to_num(raster_array_band[:, :, 0]), structure,
                    mode='same'
                ), 6
            )
        o[::, ::][np.isnan(raster_array_band[:, :, 0])] = np.nan
    elif 'nanmean' in function_variable_list[0]:
        try:
            o = np.round(
                np.divide(
                    signal.oaconvolve(
                        np.nan_to_num(raster_array_band[:, :, 0]), structure,
                        mode='same'
                    ),
                    signal.oaconvolve(
                        ~np.isnan(raster_array_band[:, :, 0]), structure,
                        mode='same'
                    )
                ), 6
            )
        except Exception as err:
            cfg.logger.log.error(str(err))
            # if scipy version < 1.4
            o = np.round(
                np.divide(
                    signal.fftconvolve(
                        np.nan_to_num(raster_array_band[:, :, 0]), structure,
                        mode='same'
                    ),
                    signal.fftconvolve(
                        ~np.isnan(raster_array_band[:, :, 0]), structure,
                        mode='same'
                    )
                ), 6
            )
        o[::, ::][np.isnan(raster_array_band[:, :, 0])] = np.nan
    elif 'nanmax' in function_variable_list[0]:
        o = maximum_filter(
            raster_array_band[:, :, 0], footprint=structure, mode='constant',
            cval=np.nan
        )
    elif 'nanmin' in function_variable_list[0]:
        o = minimum_filter(
            raster_array_band[:, :, 0], footprint=structure, mode='constant',
            cval=np.nan
        )
    elif 'median' in function_variable_list[0]:
        o = median_filter(
            raster_array_band[:, :, 0], footprint=structure, mode='constant',
            cval=np.nan
        )
    elif 'count' in function_variable_list[0]:
        try:
            o = np.round(
                signal.oaconvolve(
                    ~np.isnan(raster_array_band[:, :, 0]), structure,
                    mode='same'
                ), 6
            )
        except Exception as err:
            cfg.logger.log.error(str(err))
            # if scipy version < 1.4
            o = np.round(
                signal.fftconvolve(
                    ~np.isnan(raster_array_band[:, :, 0]), structure,
                    mode='same'
                ), 6
            )
        o[::, ::][np.isnan(raster_array_band[:, :, 0])] = np.nan
    elif 'std' in function_variable_list[0]:
        o = generic_filter(
            raster_array_band[:, :, 0], np.std, footprint=structure,
            mode='constant', cval=np.nan
        )
    elif 'percentile' in function_variable_list[0]:
        o = percentile_filter(
            raster_array_band[:, :, 0],
            percentile=int(function_variable_list[0].split(',')[1].strip(')')),
            footprint=structure, mode='constant', cval=np.nan
        )
    o[::, ::][o == cfg.nodata_val] = np.nan
    o[::, ::][np.isnan(raster_array_band[:, :, 0])] = np.nan
    o[::, ::][nodata_mask == output_no_data] = np.nan
    cfg.logger.log.debug('end')
    return [o, None]


# calculate cross rasters
def cross_rasters(*argv):
    cfg.logger.log.debug('start')
    array_function_placeholder = argv[1]
    nodata_mask = argv[2]
    function_argument = argv[7]
    function_variable = argv[8]
    _a = eval(
        function_variable.replace(
            cfg.array_function_placeholder, 'array_function_placeholder'
        )
    )
    o = np.searchsorted(function_argument, _a.ravel(), side='right').reshape(
        array_function_placeholder.shape[0],
        array_function_placeholder.shape[1]
    )
    _a = None
    if nodata_mask is not None:
        np.copyto(o, nodata_mask, where=nodata_mask[::, ::] != 0)
        stats = np.array(
            np.unique(o[nodata_mask[::, ::] == 0], return_counts=True)
        )
    else:
        try:
            o[np.isnan(array_function_placeholder[:, :, 0])] = np.nan
            stats = np.array(np.unique(o[~np.isnan(o)], return_counts=True))
        except Exception as err:
            str(err)
            try:
                stats = np.array(
                    np.unique(o[~np.isnan(o)], return_counts=True)
                )
            except Exception as err:
                cfg.logger.log.error(str(err))
                stats = None
    cfg.logger.log.debug('end')
    return [o, stats]


# calculate spectral signature
def spectral_signature(*argv):
    cfg.logger.log.debug('start')
    output_no_data = argv[0][2]
    array_function_placeholder = argv[1]
    nodata_mask = argv[2]
    # vector path
    function_argument = argv[7]
    # reference path
    function_variable = argv[8]
    temp = cfg.temp.temporary_file_path(name_suffix=cfg.tif_suffix)
    raster_vector.vector_to_raster(
        vector_path=function_argument, burn_values=1, output_path=temp,
        reference_raster_path=function_variable, extent=True
    )
    _a = raster_vector.read_raster(temp)
    _a[::, ::][nodata_mask == output_no_data] = np.nan
    array_roi = array_function_placeholder[_a == 1]
    mean = np.nanmean(array_roi)
    std = np.nanstd(array_roi)
    cfg.logger.log.debug('end')
    return [None, [mean, std]]


# get band arrays
def get_band_arrays(*argv):
    cfg.logger.log.debug('start')
    output_no_data = argv[0][2]
    array_function_placeholder = argv[1]
    nodata_mask = argv[2]
    # vector path
    function_argument = argv[7][0]
    spectral_signatures_table = argv[7][1]
    # reference path
    function_variable = argv[8]
    # iterate ROIs
    array_dictionary = {}
    for s in spectral_signatures_table.signature_id:
        vector = raster_vector.get_polygon_from_vector(
            vector_path=function_argument,
            attribute_filter="%s = '%s'" % (cfg.uid_field_name, s)
        )
        temp = cfg.temp.temporary_file_path(name_suffix=cfg.tif_suffix)
        raster_vector.vector_to_raster(
            vector_path=vector, burn_values=1, output_path=temp,
            reference_raster_path=function_variable,
            extent=True
        )
        _a = raster_vector.read_raster(temp)
        _a[::, ::][nodata_mask == output_no_data] = np.nan
        array_roi = array_function_placeholder[_a == 1]
        array_dictionary[s] = array_roi.flatten()
    cfg.logger.log.debug('end')
    return [None, array_dictionary]


# fit
def fit_classifier(*argv):
    p, temp, ram, log_process, cfg.logger = argv[0]
    classifier_list = argv[1]
    arg_dict_list = argv[2]
    result = []
    for i in range(0, len(arg_dict_list)):
        classifier = classifier_list[i]
        if log_process and log_process is not None:
            with open('%s/scikit' % temp.dir, 'w') as f:
                with redirect_stdout(f):
                    classifier.fit(**arg_dict_list[i])
        else:
            # set verbose
            try:
                classifier.verbose = 0
            except Exception as err:
                str(err)
            classifier.fit(**arg_dict_list[i])
        classifier.verbose = 0
        result.append(classifier)
    return [result, False, None]


# score
def score_classifier(*argv):
    p, temp, ram, log_process, cfg.logger = argv[0]
    classifier_list = argv[1]
    arg_dict_list = argv[2]
    result = []
    for i in range(0, len(arg_dict_list)):
        classifier = classifier_list[i]
        if log_process and log_process is not None:
            with open('%s/scikit' % temp.dir, 'w') as f:
                with redirect_stdout(f):
                    score = classifier.score(**arg_dict_list[i])
        else:
            # set verbose
            try:
                classifier.verbose = 0
            except Exception as err:
                str(err)
            score = classifier.score(**arg_dict_list[i])
        result.append(score)
    return [result, False, None]


# score classifier with stratified k fold
def score_classifier_stratified(
        process, progress_queue, argument_list, logger
):
    cfg.logger = logger
    results = []
    n = 0
    errors = False
    for d in argument_list:
        n += 1
        try:
            # calculate score by cross-validation with stratification
            scores = cross_val_score(
                d['classifier'], d['x_matrix'], d['y'],
                cv=StratifiedKFold(n_splits=5, shuffle=True, random_state=0)
            )
            score = np.array(scores)
            results.append([d['classifier'], score.mean(), score.std()])
            if progress_queue is not None:
                progress_queue.put([n, len(argument_list)], False)
        except Exception as err:
            errors = str(err)
    return [results, errors, str(process)]


# class to override numpy ufunc
class ArrayLike(np.ndarray):

    def __new__(cls, input_array):
        obj = input_array.view(cls)
        return obj

    def __mul__(self, other):
        a = self
        nan_a_mask = np.isnan(a)
        if nan_a_mask is not None:
            a = np.array(a, subok=True, copy=True)
            np.copyto(a, 1, where=nan_a_mask)
        nan_other_mask = np.isnan(other)
        if nan_other_mask is not None:
            other = np.array(other, subok=True, copy=True)
            np.copyto(other, 1, where=nan_other_mask)
        return np.multiply(a, other)

    def __rmul__(self, other):
        a = self
        nan_a_mask = np.isnan(a)
        if nan_a_mask is not None:
            a = np.array(a, subok=True, copy=True)
            np.copyto(a, 1, where=nan_a_mask)
        nan_other_mask = np.isnan(other)
        if nan_other_mask is not None:
            other = np.array(other, subok=True, copy=True)
            np.copyto(other, 1, where=nan_other_mask)
        return np.multiply(other, a)

    def __truediv__(self, other):
        nan_mask = np.isnan(other)
        if nan_mask is not None:
            other = np.array(other, subok=True, copy=True)
            np.copyto(other, 1, where=nan_mask)
        division = np.divide(self, other)
        try:
            division[nan_mask] = np.nan
        except Exception as err:
            str(err)
        return division

    def __rtruediv__(self, other):
        a = self
        nan_mask = np.isnan(self)
        if nan_mask is not None:
            a = np.array(a, subok=True, copy=True)
            np.copyto(a, 1, where=nan_mask)
        division = np.divide(other, a)
        try:
            division[nan_mask] = np.nan
        except Exception as err:
            str(err)
        return division

    def __add__(self, other):
        a = self
        nan_a_mask = np.isnan(a)
        if nan_a_mask is not None:
            a = np.array(a, subok=True, copy=True)
            np.copyto(a, 0, where=nan_a_mask)
        nan_other_mask = np.isnan(other)
        if nan_other_mask is not None:
            other = np.array(other, subok=True, copy=True)
            np.copyto(other, 0, where=nan_other_mask)
        return np.add(a, other)

    def __radd__(self, other):
        a = self
        nan_a_mask = np.isnan(a)
        if nan_a_mask is not None:
            a = np.array(a, subok=True, copy=True)
            np.copyto(a, 0, where=nan_a_mask)
        nan_other_mask = np.isnan(other)
        if nan_other_mask is not None:
            other = np.array(other, subok=True, copy=True)
            np.copyto(other, 0, where=nan_other_mask)
        return np.add(other, a)

    def __sub__(self, other):
        a = self
        nan_a_mask = np.isnan(a)
        if nan_a_mask is not None:
            a = np.array(a, subok=True, copy=True)
            np.copyto(a, 0, where=nan_a_mask)
        nan_other_mask = np.isnan(other)
        if nan_other_mask is not None:
            other = np.array(other, subok=True, copy=True)
            np.copyto(other, 0, where=nan_other_mask)
        return np.add(a, -other)

    def __rsub__(self, other):
        a = self
        nan_a_mask = np.isnan(a)
        if nan_a_mask is not None:
            a = np.array(a, subok=True, copy=True)
            np.copyto(a, 0, where=nan_a_mask)
        nan_other_mask = np.isnan(other)
        if nan_other_mask is not None:
            other = np.array(other, subok=True, copy=True)
            np.copyto(other, 0, where=nan_other_mask)
        return np.add(other, -a)



from unittest import TestCase

import remotior_sensus
from remotior_sensus.util import download_tools, files_directories


class TestDownloadTools(TestCase):

    def test_download_tools(self):
        rs = remotior_sensus.Session(
            n_processes=2, available_ram=1000, log_level=10
        )
        cfg = rs.configurations
        cfg.logger.log.debug('test')
        url = 'https://www.python.org'
        temp = cfg.temp.temporary_file_path(name_suffix='.html')
        download_tools.download_file(url=url, output_path=temp)
        self.assertTrue(files_directories.is_file(temp))
        url_2 = ''.join(
            ['https://storage.googleapis.com/gcp-public-data-sentinel-2/',
             'L2/tiles/33/S/VB/S2A_MSIL2A_20210104T094411_N0214_R036_T33S',
             'VB_20210104T122314.SAFE/GRANULE/L2A_T33SVB_A028919_20210104',
             'T094407/IMG_DATA/R60m/T33SVB_20210104T094411_B01_60m.jp2']
        )
        temp_2 = cfg.temp.temporary_file_path(name_suffix='.jp2')
        download_tools.download_file(url=url_2, output_path=temp_2)
        self.assertTrue(files_directories.is_file(temp_2))

        # clear temporary directory
        rs.close()



from unittest import TestCase

import remotior_sensus
import time


class TestProgress(TestCase):

    def test_create(self):
        rs = remotior_sensus.Session(
            n_processes=2, available_ram=1000, log_level=10
        )
        cfg = rs.configurations
        cfg.logger.log.debug('>>> test progress')

        cfg.progress.update(
            process=__name__.split('.')[-1].replace('_', ' '),
            message='starting', start=True
        )
        for e in range(0, 100):
            time.sleep(0.02)
            cfg.progress.update(
                message='test message',
                step=int(100 * (e + 1) / 100),
                percentage=int(100 * (e + 1) / 100)
            )
        cfg.progress.update(end=True)

        cfg.logger.log.debug('>>> test progress callback')
        rs.set(progress_callback=cfg.progress.print_progress)
        cfg.progress.update(
            process=__name__.split('.')[-1].replace('_', ' '),
            message='starting', start=True
        )
        for e in range(0, 100):
            time.sleep(0.02)
            cfg.progress.update(
                message='test message',
                step=int(100 * (e + 1) / 100),
                percentage=int(100 * (e + 1) / 100)
            )
        cfg.progress.update(end=True)

        # clear temporary directory
        rs.close()



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.
"""Band neighbor pixels.

This tool allows for the calculation of a function over neighbor pixels
defined by size (i.e. number of pixels) or a structure.

Typical usage example:

    >>> # import Remotior Sensus and start the session
    >>> import remotior_sensus
    >>> rs = remotior_sensus.Session()
    >>> # start the process
    >>> neighbor = rs.band_neighbor_pixels(
    ... input_bands=['file1.tif', 'file2.tif'], output_path='directory_path',
    ... size=1, circular_structure=True, stat_name='Mean', prefix='neighbor_'
    ... )
"""

from typing import Union, Optional

from remotior_sensus.core import configurations as cfg, messages
from remotior_sensus.core.bandset import BandSet
from remotior_sensus.core.bandset_catalog import BandSetCatalog
from remotior_sensus.core.output_manager import OutputManager
from remotior_sensus.core.processor_functions import raster_neighbor
from remotior_sensus.util import shared_tools


def band_neighbor_pixels(
        input_bands: Union[list, int, BandSet], size: int,
        output_path: Optional[str] = None,
        stat_name: str = None, structure: Optional[any] = None,
        circular_structure: Optional[bool] = True,
        stat_percentile: Optional[Union[int, str]] = None,
        output_data_type: Optional[str] = None,
        virtual_output: Optional[bool] = None, prefix: Optional[str] = '',
        n_processes: Optional[int] = None, available_ram: Optional[int] = None,
        bandset_catalog: Optional[BandSetCatalog] = None
) -> OutputManager:
    """Performs band neighbor pixels.

    This tool calculates a function over neighbor pixels defined by
    size (i.e. number of pixels) or structure.
    A new raster is created for each input band, where each pixel is the result
    of the calculation of the function over the neighbor pixels (e.g. the mean
    of the pixel values of a 3x3 window around the pixel).
    Available functions are:
    
        - Count
        - Max
        - Mean
        - Median
        - Min
        - Percentile
        - StandardDeviation
        - Sum

    Args:
        input_bands: input of type BandSet or list of paths or integer
            number of BandSet.
        output_path: string of output path directory or list of paths.
        size: size of dilation in pixels.
        structure: optional path to csv file of structures, if None then the
            structure is created from size.
        circular_structure: if True use circular structure.
        stat_percentile: integer value for percentile parameter.
        stat_name: statistic name as in configurations.statistics_list.
        output_data_type: optional raster output data type, if None the data type is the same as input raster.
        virtual_output: if True (and output_path is directory), save output as virtual raster of multiprocess parts.
        prefix: optional string for output name prefix.
        n_processes: number of parallel processes.
        available_ram: number of megabytes of RAM available to processes.
        bandset_catalog: optional type BandSetCatalog for BandSet number.

    Returns:
        Object :func:`~remotior_sensus.core.output_manager.OutputManager` with
            - paths = output list

    Examples:
        Perform the band neighbor of size 10 pixels with the function Sum
            >>> neighbor = band_neighbor_pixels(
            ... input_bands=['file1.tif', 'file2.tif'],
            ... output_path='directory_path',
            ... size=10, circular_structure=True, stat_name='Sum'
            ... )
    """  # noqa: E501
    cfg.logger.log.info('start')
    cfg.progress.update(
        process=__name__.split('.')[-1].replace('_', ' '), message='starting',
        start=True
    )
    # prepare process files
    (input_raster_list, raster_info, nodata_list, name_list, warped, out_path,
     vrt_r, vrt_path, n_processes,
     output_list, vrt_list) = shared_tools.prepare_process_files(
        input_bands=input_bands, output_path=output_path,
        n_processes=n_processes,
        bandset_catalog=bandset_catalog, prefix=prefix,
        multiple_output=True, virtual_output=virtual_output
    )
    stat_numpy = None
    for i in cfg.statistics_list:
        if i[0].lower() == stat_name.lower():
            stat_numpy = i[1]
            break
    cfg.logger.log.debug('stat_numpy: %s' % str(stat_numpy))
    if cfg.stat_percentile in stat_numpy:
        function_numpy = stat_numpy.replace('array', 'A')
        try:
            stat_percentile = int(stat_percentile)
            function_numpy = function_numpy.replace(
                cfg.stat_percentile, str(stat_percentile)
            )
        except Exception as err:
            cfg.logger.log.error(err)
            messages.error(str(err))
            return OutputManager(check=False)
    else:
        function_numpy = stat_numpy.replace('array', 'A, axis=2')
    cfg.logger.log.debug('function_numpy: %s' % str(function_numpy))
    if structure is None:
        if not circular_structure:
            structure = shared_tools.create_base_structure(size * 2 + 1)
        else:
            structure = shared_tools.create_circular_structure(size)
    else:
        try:
            structure = shared_tools.open_structure(structure)
        except Exception as err:
            cfg.logger.log.error(err)
            messages.error(str(err))
            return OutputManager(check=False)
    # process calculation
    n = 0
    min_p = 1
    max_p = int((99 - 1) / len(input_raster_list))
    # dummy bands for memory calculation
    dummy_bands = 3
    for i in input_raster_list:
        out = output_list[n]
        nd = nodata_list[n]
        if output_data_type is None:
            output_data_type = raster_info[n][8]
        cfg.multiprocess.run(
            raster_path=i, function=raster_neighbor,
            function_argument=structure,
            function_variable=[function_numpy], output_raster_path=out,
            output_data_type=output_data_type, output_nodata_value=nd,
            compress=cfg.raster_compression, dummy_bands=dummy_bands,
            n_processes=n_processes, available_ram=available_ram,
            boundary_size=structure.shape[0] + 1, virtual_raster=vrt_list[n],
            progress_message='processing raster %s' % str(n + 1),
            min_progress=min_p + max_p * n,
            max_progress=min_p + max_p * (n + 1)
        )
        n += 1
    cfg.progress.update(end=True)
    cfg.logger.log.info('end; neighbor pixel: %s' % str(output_list))
    return OutputManager(paths=output_list)



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.
"""
Raster to vector.

This tool allows for the conversion from raster to vector.
A new geopackage is created from the raster conversion.

Typical usage example:

    >>> # import Remotior Sensus and start the session
    >>> import remotior_sensus
    >>> rs = remotior_sensus.Session()
    >>> # start the process
    >>> vector = rs.raster_to_vector(raster_path='file.tif', output_path='vector.gpkg'))
"""  # noqa: E501

from typing import Optional

from remotior_sensus.core import configurations as cfg
from remotior_sensus.core.output_manager import OutputManager
from remotior_sensus.util import files_directories


def raster_to_vector(
        raster_path, output_path: Optional[str] = None,
        dissolve: Optional[bool] = None, field_name: Optional[str] = None,
        n_processes: Optional[int] = None, available_ram: Optional[int] = None
) -> OutputManager:
    """Performs the conversion from raster to vector.

    This tool performs the conversion from raster to vector.
    Parallel processes are used for the conversion, resulting in a vector output
    which is split as many in portions as the process numbers.
    The argument dissolve allows for merging these portions,
    but it requires additional processing time depending on vector size.

    Args:
        raster_path: path of raster used as input.
        output_path: string of output path.
        dissolve: if True, dissolve adjacent polygons having the same values;
            if False, polygons are not dissolved and the process is rapider.
        field_name: name of the output vector field to store raster values (default = DN).
        n_processes: number of parallel processes.
        available_ram: number of megabytes of RAM available to processes.

    Returns:
        object :func:`~remotior_sensus.core.output_manager.OutputManager` with
            - path = output path

    Examples:
        Perform the conversion to vector of a raster
            >>> raster_to_vector(raster_path='file.tif', output_path='vector.gpkg')
    """  # noqa: E501
    cfg.logger.log.info('start')
    cfg.progress.update(
        process=__name__.split('.')[-1].replace('_', ' '), message='starting',
        start=True
    )
    raster_path = files_directories.input_path(raster_path)
    if output_path is None:
        output_path = cfg.temp.temporary_file_path(name_suffix=cfg.gpkg_suffix)
    output_path = files_directories.output_path(output_path, cfg.gpkg_suffix)
    files_directories.create_parent_directory(output_path)
    if n_processes is None:
        n_processes = cfg.n_processes
    # perform conversion
    cfg.multiprocess.multiprocess_raster_to_vector(
        raster_path=raster_path, output_vector_path=output_path,
        field_name=field_name, n_processes=n_processes,
        dissolve_output=dissolve, min_progress=1, max_progress=100,
        available_ram=available_ram
    )
    cfg.progress.update(end=True)
    cfg.logger.log.info('end; output_path: %s' % output_path)
    return OutputManager(path=output_path)



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.

"""
Shared tools
"""

import os
import re
from typing import Union, Optional

import numpy as np

from remotior_sensus.core import configurations as cfg
from remotior_sensus.core.bandset import BandSet
from remotior_sensus.core.bandset_catalog import BandSetCatalog
from remotior_sensus.util import files_directories, raster_vector


# check input path converting to the same crs if necessary
def prepare_input_list(
        band_list, reference_raster_crs=None,
        n_processes: int = None, src_nodata=None, dst_nodata=None
):
    cfg.logger.log.debug('start')
    cfg.logger.log.debug('band_list: %s' % str(band_list))
    information_list = []
    nodata_list = []
    if reference_raster_crs is None:
        reference_raster_crs = raster_vector.get_crs(band_list[0])
    # get crs from file
    elif files_directories.is_file(reference_raster_crs):
        reference_raster_crs = raster_vector.get_crs(reference_raster_crs)
    input_list = []
    name_list = []
    warped = False
    for i in range(len(band_list)):
        name_list.append(files_directories.file_name(band_list[i]))
        crs = raster_vector.get_crs(band_list[i])
        # check crs
        same_crs = raster_vector.compare_crs(crs, reference_raster_crs)
        if not same_crs:
            t_pmd = cfg.temp.temporary_raster_path(
                name=files_directories.file_name(band_list[i]),
                extension=cfg.vrt_suffix
            )
            reference_raster = cfg.multiprocess.create_warped_vrt(
                raster_path=band_list[i], output_path=t_pmd,
                output_wkt=str(reference_raster_crs), n_processes=n_processes,
                src_nodata=src_nodata, dst_nodata=dst_nodata
            )
            warped = True
        else:
            reference_raster = band_list[i]
        input_list.append(reference_raster)
        info = raster_vector.raster_info(band_list[i])
        if info is not False:
            (gt, crs, un, xy_count, nd, number_of_bands, block_size,
             scale_offset, data_type) = info
        else:
            cfg.logger.log.error('unable to get raster info: %s', band_list[i])
            return [], [], [], [], warped
        information_list.append(
            [gt, crs, un, xy_count, nd, number_of_bands, block_size,
             scale_offset, data_type]
        )
        nodata_list.append(nd)
    cfg.logger.log.debug('end; input_list: %s' % str(input_list))
    return input_list, information_list, nodata_list, name_list, warped


# prepare process files
def prepare_process_files(
        input_bands: Union[list, int, BandSet], output_path: str,
        n_processes: Optional[int] = None,
        bandset_catalog: Optional[BandSetCatalog] = None,
        temporary_virtual_raster=True, prefix=None, multiple_output=False,
        virtual_output=None
):
    cfg.logger.log.debug('input_bands: %s' % str(input_bands))
    if n_processes is None:
        n_processes = cfg.n_processes
    # get input list
    band_list = BandSetCatalog.get_band_list(input_bands, bandset_catalog)
    # list of inputs
    (input_raster_list, raster_info, nodata_list, name_list,
     warped) = prepare_input_list(band_list, n_processes=n_processes)
    # single output path
    out_path = None
    # multiple output path list
    output_list = []
    # multiple output virtual raster check list
    vrt_list = []
    if multiple_output:
        if output_path is None:
            output_path = cfg.temp.dir
        if type(output_path) is not list and files_directories.is_directory(
                output_path
        ):
            for r in input_raster_list:
                p = os.path.join(
                    output_path, '{}{}'.format(
                        prefix, files_directories.file_name(r)
                    )
                ).replace('\\', '/')
                # check output path
                out_path, vrt_r = files_directories.raster_output_path(
                    p, virtual_output
                )
                output_list.append(out_path)
                vrt_list.append(vrt_r)
        else:
            for r in output_path:
                # check output path
                out_path, vrt_r = files_directories.raster_output_path(
                    r, virtual_output
                )
                output_list.append(out_path)
                vrt_list.append(vrt_r)
    else:
        if output_path is None:
            output_path = cfg.temp.temporary_raster_path(
                extension=cfg.vrt_suffix
            )
        # check output path
        out_path, virtual_output = files_directories.raster_output_path(
            output_path, virtual_output
        )
    # create virtual raster of input
    if temporary_virtual_raster:
        temporary_virtual_raster = \
            raster_vector.create_temporary_virtual_raster(
                input_raster_list
            )
    return [input_raster_list, raster_info, nodata_list, name_list, warped,
            out_path, virtual_output,
            temporary_virtual_raster, n_processes, output_list, vrt_list]


# create base structure
def create_base_structure(size: int):
    structure = np.ones((size, size))
    return structure


# create a circular structure
def create_circular_structure(radius):
    circle = np.zeros([radius * 2 + 1, radius * 2 + 1])
    for x in range(radius * 2 + 1):
        for y in range(radius * 2 + 1):
            if (x - radius) ** 2 + (y - radius) ** 2 <= radius ** 2:
                circle[x, y] = 1
    return circle


# open structure file
def open_structure(structure):
    text = open(structure, 'r')
    lines = text.readlines()
    c = None
    for b in lines:
        b = b.replace('nan', 'np.nan').replace(cfg.new_line, '')
        a = eval(b)
        i = np.array(a, dtype=np.float32)
        try:
            c = np.append(c, [i], axis=0)
        except Exception as err:
            str(err)
            try:
                c = np.append([c], [i], axis=0)
            except Exception as err:
                str(err)
                c = i
    return c


# expand list of lists
def expand_list(input_list):
    expanded_list = []
    for c in input_list:
        expanded_list.extend(c)
    return expanded_list


# join path
def join_path(*argv):
    path = os.path.join(*argv)
    return path


# replace ignoring case
def replace(input_string: str, old_value: str, new_value: str):
    old_value = old_value.replace('*', '\\*').replace(
        '(', '\\('
    ).replace(')', '\\)')
    output = re.sub(old_value, new_value, input_string, flags=re.IGNORECASE)
    return output



# Remotior Sensus , software to process remote sensing and GIS data.
# Copyright (C) 2022-2023 Luca Congedo.
# Author: Luca Congedo
# Email: ing.congedoluca@gmail.com
#
# This file is part of Remotior Sensus.
# Remotior Sensus is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by 
# the Free Software Foundation, either version 3 of the License,
# or (at your option) any later version.
# Remotior Sensus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty 
# of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Remotior Sensus. If not, see <https://www.gnu.org/licenses/>.
"""
Raster report.

This tool allows for the calculation of a report providing information 
extracted from a raster.

Typical usage example:

    >>> # import Remotior Sensus and start the session
    >>> import remotior_sensus
    >>> rs = remotior_sensus.Session()
    >>> # start the process
    >>> report = rs.raster_report(raster_path='file.tif', output_path='report.csv')
"""  # noqa: E501

import io
from typing import Optional

import numpy as np

from remotior_sensus.core import configurations as cfg, table_manager as tm
from remotior_sensus.core.output_manager import OutputManager
from remotior_sensus.core.processor_functions import (
    raster_unique_values_with_sum
)
from remotior_sensus.util import (
    files_directories, raster_vector, read_write_files
)


def raster_report(
        raster_path: str, output_path: Optional[str] = None,
        nodata_value: Optional[int] = None,
        n_processes: Optional[int] = None, available_ram: Optional[int] = None
):
    """Calculation of a report providing information extracted from a raster.

    This tool allows for the calculation of a report providing information
    such as pixel count, area per class and percentage of the total area.
    The output is a csv file.
    This tool is intended for integer rasters.

    Args:
        raster_path: path of raster used as input.
        output_path: string of output path.
        nodata_value: value to be considered as nodata.
        n_processes: number of parallel processes.
        available_ram: number of megabytes of RAM available to processes.

    Returns:
        object :func:`~remotior_sensus.core.output_manager.OutputManager` with
            - path = output path

    Examples:
        Perform the report of a raster
            >>> raster_report(raster_path='file.tif', output_path='report.csv')
    """
    cfg.logger.log.info('start')
    cfg.progress.update(
        process=__name__.split('.')[-1].replace('_', ' '), message='starting',
        start=True
    )
    raster_path = files_directories.input_path(raster_path)
    if output_path is None:
        output_path = cfg.temp.temporary_file_path(name_suffix=cfg.csv_suffix)
    output_path = files_directories.output_path(output_path, cfg.csv_suffix)
    files_directories.create_parent_directory(output_path)
    (gt, crs, crs_unit, xy_count, nd, number_of_bands, block_size,
     scale_offset, data_type) = raster_vector.raster_info(raster_path)
    pixel_size_x = abs(gt[1])
    pixel_size_y = abs(gt[5])
    if n_processes is None:
        n_processes = cfg.n_processes
    # dummy bands for memory calculation
    dummy_bands = 2
    # multiprocess calculate unique values and sum
    cfg.multiprocess.run(
        raster_path=raster_path, function=raster_unique_values_with_sum,
        use_value_as_nodata=nodata_value, n_processes=n_processes,
        available_ram=available_ram, keep_output_argument=True,
        dummy_bands=dummy_bands,
        progress_message='unique values', min_progress=2, max_progress=99
    )
    cfg.progress.update(message='output table', step=99)
    # calculate sum of values
    cfg.multiprocess.multiprocess_sum_array(nodata_value)
    unique_val = cfg.multiprocess.output
    # create table
    table = _report_table(
        table=unique_val, crs_unit=crs_unit, pixel_size_x=pixel_size_x,
        pixel_size_y=pixel_size_y
    )
    # save combination to table
    read_write_files.write_file(table, output_path)
    cfg.progress.update(end=True)
    cfg.logger.log.info('end; raster report: %s' % output_path)
    return OutputManager(path=output_path)


def _report_table(table, crs_unit, pixel_size_x, pixel_size_y):
    """Create text for tables."""
    cfg.logger.log.debug('start')
    total_sum = table['sum'].sum()
    text = []
    cv = cfg.comma_delimiter
    nl = cfg.new_line
    # table
    if 'degree' not in crs_unit:
        output_field_names = ['RasterValue', 'PixelSum', 'Percentage %',
                              'Area [%s^2]' % crs_unit]
        input_field_names = ['new_val', 'sum', 'percentage', 'area']
        cross_class = tm.calculate_multi(
            matrix=table, expression_string_list=[
                '"sum" * %s * %s' % (pixel_size_x, pixel_size_y),
                '100 * "sum" / %s' % total_sum],
            output_field_name_list=['area', 'percentage'],
            progress_message=False
        )
    else:
        output_field_names = ['RasterValue', 'PixelSum', 'Percentage %',
                              'Area not available']
        input_field_names = ['new_val', 'sum', 'percentage', 'area']
        # area is set to nan
        cross_class = tm.calculate_multi(
            matrix=table,
            expression_string_list=['100 * "sum" / %s' % total_sum,
                                    'np.nan * "sum"'],
            output_field_name_list=['percentage', 'area'],
            progress_message=False
        )
    redefined = tm.redefine_matrix_columns(
        matrix=cross_class, input_field_names=input_field_names,
        output_field_names=output_field_names, progress_message=False
    )
    # create stream handler
    stream1 = io.StringIO()
    np.savetxt(stream1, redefined, delimiter=cv, fmt='%1.2f')
    matrix_value = stream1.getvalue()
    for c in output_field_names:
        text.append(c)
        text.append(cv)
    text.pop(-1)
    text.append(nl)
    text.append(matrix_value.replace('.00', ''))
    text.append(nl)
    joined_text = ''.join(text)
    return joined_text
